
import sam.lang.antlr.* ;
import sam.lang.* ; 
import sam.infer.* ; 
import sam.util.* ; 
import java.io.* ; 
import java.net.* ; 
import java.awt.* ;
import java.util.* ;

/**
  *  colog 
  *  Main-class in cologX.jar
  *  @version 11 January 2015
  *  No fol loads this version
  *  @author &copy; John R. Fisher
  */
public class colog { 
   /**
     * Allow main to throw exceptions.
      */
   public static void main(String[] args) throws Exception {  
      boolean block = false ; 
      int num = 0 ; 
      try {
         num = Integer.valueOf(args[0]) ;
      }
      catch (Exception a) { ; }   // 
      // Start numbered GUI 
      if (args.length == 1 && num > 0) {  // numbered colog process 
         try {
            final sam.gui.Colog co = new sam.gui.Colog(Integer.valueOf(args[0])) ; 
            block = true ;
          }
         catch (Exception xx) { System.out.println(xx) ; } 
      }
      if (!block &&args.length > 0) {  // command-line driver 
         if (args[0].equals("help") || args[0].equals("h")  || args[0].equals("version") || args[0].equals("v") || args[0].equals("?")) { 
            System.out.println(CologConstants.HELP) ;  
            return ; 
         }  
         ///// - command-line search  - ///////////////////////////////// 
         File desc = null ; 
         URL url = null ;
         int depth = 300 ; 
         int width = 1000 ; 
         int complexity = -1 ;  // OFF 
         boolean log = false ;
         // parse comline ...
         for (int a = 0 ; a < args.length ; a++) {
            if (args[a].startsWith("file=") && url==null)   // ignore if url already specified
                  desc = new File(args[a].substring(5).trim()) ;  
            else if (args[a].startsWith("url=") && desc==null) // ignore if file already specified
                 // allow java.net.MalformedURLException to stop process
                  url = new URL(args[a].substring(4).trim()) ;  
            else if (args[a].startsWith("depth=")) { 
               depth = Integer.parseInt(args[a].substring(6)) ;  
               //System.out.println("depth="+depth) ; 
            }
            else if (args[a].startsWith("width=")) {
               width = Integer.parseInt(args[a].substring(6)) ;  
               //System.out.println("width="+width) ; 
            }
            else if (args[a].startsWith("cut=")) 
               complexity = Integer.parseInt(args[a].substring(4)) ;  
            else if (args[a].startsWith("log=")) { 
               String v = args[a].substring(4).trim() ;
               log = (v.equals("true") ? true : false) ;  
            }
         }
         // System.out.println("desc="+desc+"  "+"url="+url) ; 
         // require good file description
         if (desc != null && !desc.exists()) {
            System.out.println("no such file") ; return ;
         }  
         // require good file description
         if (desc == null && url == null ) {
            System.out.println("no input theory") ; return ;
         }
         // otherwise .. OLD ...
         //sam.infer.Tree tree = new sam.infer.Tree(depth,width) ;  
         //tree.complexity_limit = complexity ; 
         //tree.setLogging(log) ; 

         if (url != null) {
            boolean loaded = true ; // but ...
            Theory theory = null ; 
            try { 
               String sf = url.getFile() ; 
               theory = open_geolog_url(url) ; 
            } 

            catch (Exception xx) {  
               loaded = false ;  
               System.out.println("Cannot load "+url.toString()) ;
            }
            if (loaded) { 
               width = (theory.isFlat() ? 1 : width) ;  // no disjunction?
               sam.infer.Tree tree = newTree(theory,depth,width,complexity,log) ;
               System.out.println(search("url=",url.toString(),theory,tree,log)) ; 
            } else { return ; }
            // ignore file methods below this point 
            return ;
         }

         File[] fileList ; 
         if (desc != null && desc.isDirectory()) fileList = desc.listFiles(new Visible()) ;
         else { // desc is a theory file
            fileList = new File[1] ;
            fileList[0] = desc ;
         }
         for (int t = 0 ; t < fileList.length ; t++) { 
            boolean loaded = true ; // but ...
            Theory theory = null ; 
            try { 
               String sf = fileList[t].getName() ; 
               theory = open_geolog_file(fileList[t]) ; 
            } 
            catch (Exception xx) {  
               loaded = false ;  
               System.out.println("Cannot load "+fileList[t].getName()) ;
            }
            if (loaded) { 
               int wtmp = (theory.isFlat() ? 1 : width) ;  // no disjunction?
               sam.infer.Tree tree = newTree(theory,depth,wtmp,complexity,log) ;
               System.out.println(search("file=",fileList[t].getName(),theory,tree,log)) ; 
            } 
         } // end for all files
      }  // end command line
      else if (!block) { // GUI interaction, no input file
         new sam.gui.Colog(0) ; 
      } // end GUI launch
   }

   /**
     *  Encapsulate a proof search for a theory.
     *  @return command-line results 
      */
   public static String search(String prefix, String fname,Theory theory,Tree tree,boolean log) {
      long time =  System.currentTimeMillis() ; // milliseconds 
      tree.dfSearch() ;  
      time = System.currentTimeMillis() - time ; 
      // report on the run 
      String message = "" ; 
      if (tree.counterModels.isEmpty() && tree.tooDeepBranches.isEmpty() && tree.tooWideBranches.isEmpty()) 
         message = "PROOF" ; 
      else if (!tree.counterModels.isEmpty())  {
         String s = (tree.counterModels.size() == 1 ? "" : "s") ; 
         if (tree.complexity_limit > -1) 
            message = "POSSIBLE_COUNTER_MODEL"+s ; 
         else
            message = "COUNTER_MODEL"+s ; 
      }
      else message = "BLOCKED_BRANCH" + 
          "{" + (!tree.tooDeepBranches.isEmpty() ? "D" : "") + 
                (!tree.tooWideBranches.isEmpty() ? "W" : "") +"}" ; 
      StringBuffer sb = new StringBuffer() ;
      //sb.append("colog: "+prefix) ;
      sb.append(fname) ;
      sb.append(", ") ;
      sb.append("d=") ;
      sb.append(tree.depth) ;
      sb.append(", ") ;
      sb.append("w=") ;
      sb.append(tree.width) ;
      sb.append(", ") ;
      sb.append("c=") ;
      sb.append(tree.complexity_limit) ;
      sb.append(", ") ;
      sb.append(message) ;
      sb.append(", time=") ;
      sb.append(String.valueOf(time)) ;
      sb.append("ms") ; 
      sb.append(",\n   inferences=") ;
      sb.append(String.valueOf(tree.inferenceTally)) ;
      sb.append(", facts=") ;
      sb.append(String.valueOf(tree.count)) ;
      sb.append(", branches=") ;
      sb.append(String.valueOf(tree.numberOfBranches())) ; 
      sb.append(", deepest=") ;
      sb.append(String.valueOf(tree.descent)) ; 
      sb.append(", \n   subsumed=") ;
      sb.append(String.valueOf(tree.subsumed)) ;
      sb.append(", irrelevant_branches=") ;
      sb.append(String.valueOf(tree.unused_branch_count)) ;
      sb.append(", goals=") ;                 //  2/10/2016
      sb.append(String.valueOf(tree.goals)) ; //  2/10/2016
      if (log) sb.append("\n"+tree.proof() + "----------\n") ;
      return sb.toString() ; 
   }

   // filter visible files in a directory
   static class Visible implements FileFilter {
      public boolean accept(File f) { 
         return !f.isHidden() ;
      }
   }


   /*** Open a colog theory from file */
   protected static Theory open_geolog_file(File f) { 
      Theory theory = null ; // null return if exception
      try { 
         FileInputStream in = new FileInputStream(f) ;
         cologLoader loader = new cologLoader(in) ;    
         theory = loader.theory ;  
      }
      catch (Exception x) { System.out.println(x) ; }
      return theory ; 
   }


   /*** Open a colog theory from URL */
   protected static Theory open_geolog_url(URL url) { 
      Theory theory = null ; // null return if exception
      try { 
         InputStream in = url.openStream() ; 
         cologLoader loader = new cologLoader(in) ;    
         theory = loader.theory ;  
      }
      catch (Exception x) { System.out.println(x) ; }
      return theory ; 
   }


   // construct Tree using Theory hints 
   private static Tree newTree(Theory theory, int d, int w, int c, boolean log) {
      // No hints, use defaults or comline values
      // Must remove hints to override their values!
      if (theory.hints == null) {
         sam.infer.Tree tree = new sam.infer.Tree(d,w) ;  
         tree.setTheory(theory) ; 
         tree.complexity_limit = c ; 
         tree.setLogging(log) ; 
         return tree ; 
      }
      else { 
         // defaults or comline values are automatically overridden by hints
         int depth = d,
             width = w,
             cut = c ; 
         Iterator it = theory.hints.keySet().iterator() ;
         while (it.hasNext()) { 
            String attr = (String)it.next() ; 
            if (attr.equals("cut")) { 
               try { 
                  int c1 = Integer.decode(theory.hints.get(attr)) ; 
                  if (c1 >= 0) { // N.B. cut=0 allows constants (including Skolems) but no functions(-..)
                     cut = c1 ; 
               }
               else cut = 0 ;
               } catch (Exception cx) {  System.out.println("Bad hint for cut: "+ cx) ; }  // cannot parse int
            } // end parse cut
            else if (attr.equals("width")) { 
               try { 
                  int w1 = Integer.decode(theory.hints.get(attr)) ; 
                  if (w > 0) width = w1 ;
               }
               catch (Exception wx) { System.out.println("Bad hint for width: "+ wx) ;}  // not a width 
            } // end parse width
            else if (attr.equals("depth")) { 
               try { 
                  int d1 = Integer.decode(theory.hints.get(attr)) ; 
                  if (d > 0) depth = d1 ;
               }
               catch (Exception dx) { System.out.println("Bad hint for width: "+ dx) ;}  // not a depth 
            } // end parse depth
            //System.out.println(attr+"="+this.theory.hints.get(attr)) ;
         } // end while
         sam.infer.Tree tree = new sam.infer.Tree(depth,width) ;  
         tree.setTheory(theory) ; 
         tree.complexity_limit = cut ; 
         tree.setLogging(log) ; 
         return tree ; 
      }
   }

}
