package sam.fol ; 

import org.antlr.runtime.* ;
import sam.fol.antlr.* ; 
import sam.fol.* ; 
import java.util.* ; 
import java.io.* ; 
import java.net.* ;
import java.awt.datatransfer.* ; 
import javax.swing.* ;
import javax.swing.event.* ;

/**
  *  Process TPTP inputs, convert to fol form.
  * @version 20 April2010  
  * @author &copy; John Fisher 
  */
public class TPTP {  

   public Vector <TPTPrecord> theory ; 

   /**
     * Create empty TPTP database.
     */
   public TPTP() { 
      this.theory = new Vector <TPTPrecord> () ; // generic nonsense
   }

   /**
     * Loading error, test for...
     */
   public boolean error ; 

   /**
     * Load TPTP from a file.  NOT FINISHED
     */
   public TPTP(File file) { 
      this.theory = new Vector <TPTPrecord> () ; 
      StringBuffer buffer = new StringBuffer() ; 
      boolean error = false ; 
      try { 
         FileReader fr = new FileReader(file) ;
         BufferedReader br = new BufferedReader(fr) ;
         //StringBuffer sbuffer = new StringBuffer() ; // efficiency!
         String cr = "\n" ; 
         String s = "" ;
         while (s != null) {
            s = br.readLine() ;
            if (s != null) { 
               buffer.append(s) ; 
               buffer.append(cr) ; 
            }
         }
         br.close() ; 
      }
      catch(Exception e) { 
         System.out.println(e) ; 
         error = true ;  
      }
      if (!error) {
         remove(buffer,"$") ; // excise "$"'s
         String mode = (buffer.indexOf("cnf")>=0 ? "cnf" : "fof") ; 
         //System.out.println(mode) ;      
         int begin = buffer.indexOf(mode), end ; 
         while (begin > -1) { 
            begin = buffer.indexOf(mode,begin) ;
            end = buffer.indexOf(".",begin) ; 
            int p = buffer.indexOf("(",begin) ; 
            int c1 = buffer.indexOf(",",p) ; 
            int c2 = buffer.indexOf(",",c1+1) ;  
            String annotation = buffer.substring(p+1,c1) ; 
            begin = buffer.indexOf(mode,end) ;
            String role = buffer.substring(c1+1,c2) ; 
            String formula = buffer.substring(c2+1,buffer.indexOf(".",c2)) ; 
            formula = formula.substring(0,formula.lastIndexOf(")") ) ; 
////////////
            // bothersome modifications...
            if (!topLevel(formula)) { // trim outer parens
               int x = formula.indexOf("("), 
                   y = formula.lastIndexOf(")") ; 
               formula = formula.substring(x+1,y) ; 
            }
           
            if (malformed(formula)) {   // detect ~(! ...) or ~(? ...) and trim parens 
               int x = formula.indexOf("("), 
                   y = formula.lastIndexOf(")") ; 
               formula = formula.substring(0,x-1) + formula.substring(x+1,y) ; 
            }
/////////////
            this.theory.add(new TPTPrecord(role,mode,annotation,formula)) ; 
         }
      }
   }

    // detect ~(! ...) or ~(? ...) -- a little expensive but necessary
   private static boolean malformed(String w) { 
      String s = w.trim() ; 
      if (!s.startsWith("~")) return false ; // otherwise 
      s = s.substring(1).trim() ; 
      if (!s.startsWith("(")) return false ; // otherwise .. 
      s = s.substring(1).trim() ; 
      if ( s.startsWith("!") || s.startsWith("?") ) return true ;
      return false ;
   }

   protected void remove(StringBuffer b, String s) { 
      boolean flag = true ;
      while (flag) { 
         int p = b.indexOf(s) ; 
         if (p > -1) b.deleteCharAt(p) ; 
         else flag = false ; 
      }
   }

   /**
     *  TPTP toString() displays ANTLR readable fol input.
     */
   public String toString() { 
      Iterator records = this.theory.iterator() ; 
      StringBuffer b = new StringBuffer() ; 
      while (records.hasNext()) {
         b.append(records.next().toString()) ; 
         b.append("\n") ; 
         b.append("\n") ;           
      }
      return b.toString() ; 
   }

   /**
     * Is '&', '|', '=>' or '<=>' at top level of a wff string?
     * That is, does the left parens capture an
     * occurrence of &,|,=>,<=> ?
     */
   public static boolean topLevel(String w) { 
      int a = w.indexOf("(") + 1 ;  // first after
      //System.out.println(w) ; 
      int balance = 0 ; 
      while (a < w.length()) {
         //System.out.print(w.charAt(a)) ; 
         if (w.startsWith("(",a) ) { balance-- ; a++; } 
         else if (w.startsWith(")",a) ) { balance++ ; a++ ; } 
         else if (w.startsWith("&",a)) { 
            if (balance == 0) return true ; 
            else a++ ;
         }
         else if (w.startsWith("|",a)) { 
            if (balance == 0) return true ; 
            else a++ ;
         }
         else if (w.startsWith("<=>",a)) { 
            if (balance == 0) return true ; 
            else a=a+3 ;
         }
         else if (w.startsWith("=>",a)) { 
            if (balance == 0) return true ; 
            else a=a+2 ; 
         }
         else a++ ; 
      }
      return false ; 
   }
   
   /**
     *  File(TPTP format) ==> File(Fol format), 
     *  @return Fol File. 
     */
   public static File TPTP2Fol(File f) {
      File f2 = new File(f.getAbsolutePath() + ".fol") ; 
      TPTP tptp = new TPTP(f) ; 
      String s = tptp.toString() ; 
      try { 
         FileWriter fw = new FileWriter(f2) ; 
         fw.write(s,0,s.length()) ; 
         fw.close() ; 
      } catch (Exception x) { System.out.println(x) ; }
      return f2 ; 
   }

   /**
     *  Convert File from command line.
     */
   public static void main(String[] args) { 
      System.out.println(TPTP2Fol(new File(args[0])) + " created.") ; 
   }

   public class TPTPrecord { 
      public String role,        // e.g. axiom, conjecture, negated_conjecture, ...
                    mode,        // e.g. fof, cnf, fol, ...
                    annotation,  // describes the input
                    formula ;    // text form of input logical formula
      
      public TPTPrecord(String r,String m,String a,String f) { 
         this.role = r ; 
         this.mode = m ; 
         this.annotation = a ; 
         this.formula = f ; 
      }

      /**
        * Strip off the last ')' & '.' 
        */
      String strip(String w) { 
         StringBuffer b = new StringBuffer(w) ; 
         int p = b.indexOf(".") ; 
         b.deleteCharAt(p) ; 
         p = b.lastIndexOf(")") ; 
         b.deleteCharAt(p) ; 
         return b.toString() ;
      }

      public String toString() { 
         StringBuffer b = new StringBuffer() ;
         b.append(role) ; b.append(':') ; 
         b.append(mode) ; b.append(':') ; 
         b.append(annotation) ; b.append(':') ; 
         b.append("\n") ; 
         b.append(formula) ; b.append('.') ; 
         b.append("\n") ; 
         return b.toString() ; 
      }

   } // end TPTPrecord class definition

} // end TPTP class
