
package sam.fol ; 

import sam.lang.* ; 
import java.util.* ; 
import java.awt.* ;  // for Point

import java.io.* ;           // 11.1.2014  
import sam.lang.antlr.* ;    // 11.1.2014

// Typed quantifiers
// Consistency rules only for negated atoms in consequent of geolog rule
// auto dom for quantifier types
// No auto $function closures (search: NOCLOSE$)
// check quantifier scoping
// Herbrandization for coherent rule goals

/**
  * A WFF is a pseudo-tree of logical forms.
  * Coherent Fore-and-aft Translations  
  * @version 1 November 2014  (see note at bottom of source) 
  * @author &copy; John Fisher 
  * Communication with translator is via method parameter (11/5/2012).
  * 
  */
public class WFF { 

   // strings designating fore/aft translation
   public static String FORE = "FORE" ;    
   public static String AFT  = "AFT" ;  

   //////////////////////////
   // types of wff
   //////////////////////////
   public static int FUNCTOR = 0 ; //    w
   public static int AND = 1 ;     //    w    w ...
   public static int OR = 2 ;      //    w    w ...
   public static int NOT = 3 ;     //    w
   public static int IF = 4 ;      //    w    w
   public static int IFF = 5 ;     //    w    w
   public static int ALL = 6 ;     //   vars  w
   public static int EXISTS = 7 ;  //   vars  w

   //////////////////////////////////////
   //     op      operands
   //  FUNCTOR: [functor], <br>
   //  AND: [wff,wff,..], <br>
   //  OR: [wff,wff,..], <br>
   //  NOT: [wff] , <br>
   //  IF: [wff,wff] , <br>
   //  IFF: [wff,wff] , <br>
   //  ALL: [vars,wff] , <br>
   //  EXISTS: [vars,wff] , <br>  
   /////////////////////////////////////// 
   public int op ; 
   public Vector<Object> operands ;  // multi class 
   protected static int hCount = 0 ; 

   // The universal domain name
   private static String DOM = "dom" ;  

   /**
     *  (partially) Construct WFF knowing 
     *  logical type (AND/OR/NOT/...).
     *  Then add the operands.  
     */
   public WFF(int t) { 
      this.op = t ;
      this.operands = new Vector<Object>() ;
   }

   /**
     *  toString()
     */
   public String toString() { 
      StringBuffer s = new StringBuffer() ; 
      if (this.op == 0) { // FUNCTOR  -- needed modification
          s.append(((Functor)this.operands.elementAt(0)).toString()) ;
      }
      else if (this.op == 1) {  //AND
         s.append("(") ; 
         Iterator it = this.operands.iterator() ; 
         s.append(((WFF)it.next()).toString()) ;
         while (it.hasNext()) { 
            s.append(" & ") ;
            s.append(((WFF)it.next()).toString()) ; 
         }
         s.append(")") ; 
      }
      else if (this.op == 2) {  //OR
         s.append("(") ; 
         Iterator it = this.operands.iterator() ; 
         s.append(((WFF)it.next()).toString()) ;
         while (it.hasNext()) { 
            s.append(" | ") ;
            s.append(((WFF)it.next()).toString()) ; 
         }
         s.append(")") ; 
      }
      else if (this.op == 3) {  //NOT
         s.append("~") ; 
         s.append(((WFF)this.operands.elementAt(0)).toString()) ;
      }
      else if (this.op == 4) {  //IF
         s.append("(") ; 
         s.append(((WFF)this.operands.elementAt(0)).toString()) ;
         s.append(" => ") ; 
         s.append(((WFF)this.operands.elementAt(1)).toString()) ;
         s.append(")") ; 
      }
      else if (this.op == 5) {  //IFF
         s.append("(") ; 
         s.append(((WFF)this.operands.elementAt(0)).toString()) ;
         s.append(" <=> ") ; 
         s.append(((WFF)this.operands.elementAt(1)).toString()) ;
         s.append(")") ; 
      }
      else if (this.op == 6) {  //ALL
         s.append("![") ; 
         Iterator vars = ((Vector)this.operands.elementAt(0)).iterator() ;
         s.append(((TypedVariable)(vars.next())).toFolString()) ;
         while(vars.hasNext()) { 
            s.append(",") ;
            s.append(((TypedVariable)(vars.next())).toFolString()) ; 
         }
         s.append("]:") ; 
         s.append(((WFF)this.operands.elementAt(1)).toString()) ;
      }
      else if (this.op == 7) {  //EXISTS
         s.append("?[") ; 
         Iterator vars = ((Vector)this.operands.elementAt(0)).iterator() ;
         s.append(((TypedVariable)(vars.next())).toFolString()) ;
         while(vars.hasNext()) { 
            s.append(",") ;
            s.append(((TypedVariable)(vars.next())).toFolString()) ; 
         }
         s.append("]:") ; 
         s.append(((WFF)this.operands.elementAt(1)).toString()) ;
      }   
      return s.toString() ;
   }

   /**
     *  @return logical negation of this WFF
      */
   public WFF negation() {
      WFF nwff = new WFF(NOT) ; 
      nwff.operands.add(this) ; 
      return nwff ;
   }


///////////////////////////////////////////////////////////////
///////// CONVERT TO GEOLOG RULES
/////////   coherent rules, domain declarations and closures

   /**
     *  Convert to geolog theory, 
     *  positive orientation.
      */
   public void toGeolog(Theory domain,Theory closure,Theory thy,Theory consistency,TreeSet<String> guards,String method) { 
      HashSet<TypedVariable> varSet = new HashSet<TypedVariable>() ;  // collect universal variables
      int check = this.checkGoal(varSet) ; // what kind of goal am I?
      // This WFF is modified if it can be Herbrandized,
      // and the WFF will print with new constants instead
      // of the original variables.
      //   =0, not coherent form goal
      //   =1, (all X in varSet)<conjunction of functors>
      //   =2, (all X in varSet)<if ... then ...>
      //   =3, (all X in varSet)<disjunction of existential forms>
      if (check == 0) {  // not coherent form goal
         // assert reified non-goal wff (+)
         VarList varlist = new VarList() ; 
         Rule rule = makeAssertion(this,varlist) ;
         thy.add(rule) ; 
         // now translate ...
         this.pAft(domain,closure,thy,consistency,varlist,guards,method) ;  
      }
      else {  // we have an allowed goal form ...
         hCount++ ;  
         WFF w = (WFF)this.operands.elementAt(0) ;  // the goal form itself
         // silent move past any ALL's (the variables are already in varSet)
         while (w.op == ALL) w = (WFF)w.operands.elementAt(1) ;
         // assert types for non-DOM Herbrand variables
         if (!varSet.isEmpty()) {  // there are universal variables
            Iterator it = varSet.iterator() ;   
            Conjunction b = new Conjunction() ; 
            boolean flag = false ; // whether or not there are user types
            while (it.hasNext()) { 
               TypedVariable tv = (TypedVariable)it.next() ; 
               if (!tv.typename.equals(DOM)) {
                  guards.add(tv.typename) ; 
                  flag = true ; 
                  Functor f = new Functor(tv.typename,1,Functor.FCNL) ; 
                  f.args[0] = new Functor("@"+tv.name+hCount,0,Functor.FCNL) ; 
                  b.add(f) ; 
               }
            }
            if (flag) {  // at least 1 non-DOM 
               Functor tru = new Functor("true",0,Functor.FCNL) ; 
               Conjunction a = new Conjunction() ; 
               a.add(tru) ; 
               Disjunction d = new Disjunction() ; 
               d.add(b) ; 
               Rule rule = new Rule(a,d) ; 
               domain.add(rule) ; 
            }
         }  
         WFF newgoal = new WFF(IF) ;
         Functor gol = new Functor("goal",0,Functor.FCNL) ; 
         WFF newgoalc = new WFF(FUNCTOR) ; 
         newgoalc.operands.add(gol) ;
         if (check == 1 || check == 3) { // N.B. 3 >> 1 as special case
            // e.g., w ~ (p(X)&q(X)),   w ~ (?[Y]:r(X,Y) | ?[Z]:s(X,Z))
            newgoal.operands.add(w) ;
            newgoal.operands.add(newgoalc) ; 
            herbrandize(newgoal,varSet) ;
            VarList varlist = new VarList() ;
            Rule rule = makeAssertion(newgoal,varlist) ; 
            thy.add(rule) ;
            newgoal.pAft(domain,closure,thy,consistency,varlist,guards,method) ; 
         }
         else if (check == 2) {    // e.g., w ~ (p(X)&q(X) => (?[Y]:r(X,Y) | ?[Z]:s(X,Z))).
            // peel off antecedent and consequent, ant => cons
            WFF ant =  (WFF)w.operands.elementAt(0) ;
            WFF cons =  (WFF)w.operands.elementAt(1) ;  
            herbrandize(ant,varSet) ; 
            newgoalc.operands.add(gol) ;
            newgoal.operands.add(cons) ; 
            newgoal.operands.add(newgoalc) ; 
            herbrandize(newgoal,varSet) ; 
            VarList varlist = new VarList() ; 
            Rule rule1 = makeAssertion(newgoal,varlist) ;
            thy.add(rule1) ; 
            newgoal.pAft(domain,closure,thy,consistency,varlist,guards,method) ; 
            varlist = new VarList() ; 
            Rule rule2 = makeAssertion(ant,varlist) ;
            thy.add(rule2) ; 
            ant.pAft(domain,closure,thy,consistency,varlist,guards,method) ; 
         }
      }
   } // end toGeolog


///////////////////////////////////////////////////////////////////////// 10.1.2011

   /**
     *  Recognize a coherent goal.
     *  @return 0 if not coherent goal, otherwise coherent and ...
     *  @return 1 if universalized conjunction of atoms
     *  @return 2 if universalized implicational form:
     *  @return 3 if universalized disjunction of existential conjunctive forms
     *  <pre>
     *    ( ![..]:* ( (a & ..) => ( ?[..]:* (c & ..) |.. ) => goal ).
     *                  ^isAndForm              ^isOrForm
     *                < 1 >   > 2 <       < 3 >
     *  </pre>
     *  The quantifiers can be multiple. 
     *  Store the universal variables in varSet.
     */
   protected int checkGoal(HashSet<TypedVariable> varSet) {   
      // check for (.. => goal) form first
      if (this.op != IF) return 0; 
      WFF w = (WFF)this.operands.elementAt(1) ; 
      if (w.op != FUNCTOR) return 0 ; 
      Functor f = (Functor)w.operands.elementAt(0) ; 
      if (!f.name.equals("goal")) return 0 ; 
      ////////// outer form OK (.. =>goal), now check further
      w = (WFF)this.operands.elementAt(0) ;  // the possible coherent form
      // move past ALL's, collecting the variables
      while (w.op == ALL) { 
         Iterator vars = ((Vector)w.operands.elementAt(0)).iterator() ;
         while (vars.hasNext()) 
            varSet.add((TypedVariable)vars.next()) ;
         w = (WFF)w.operands.elementAt(1) ;
      }
      ///>>> 1?
      if (w.isAndForm()) return 1 ;
      ///>>> 2?
      else if (w.op == IF) { 
         // check the antecedent
         WFF a = (WFF)w.operands.elementAt(0) ; 
         if (!a.isAndForm()) return 0 ;
         // check consequent 
         WFF c = (WFF)w.operands.elementAt(1) ;  
         if (!c.isOrForm()) return 0 ;
         else return 2 ; 
      }
      else if (w.isOrForm()) return 3 ;
      else return 0 ; 
   }

   /**
     *  @return whether or not this WFF is a conjunction of atoms (FUNCTOR)
     */
   protected boolean isAndForm() { 
      if (this.op == FUNCTOR) return true ; // conjunction of one
      if (this.op != AND) return false ; 
      // otherwise check conjuncts to see whether they are FUNCTOR
      Iterator it = this.operands.iterator() ;
      while (it.hasNext()) 
         if (((WFF)it.next()).op != FUNCTOR) return false ;
      return true ; 
   }
      
   /**
     *  @return whether or not this WFF is a disjunction of existential conjunctions
     */
   protected boolean isOrForm() { 
      if (this.isExistsForm()) return true ; // one disjunct
      else if (this.op == OR) {
         Iterator it = this.operands.iterator() ; 
         while (it.hasNext()) 
            if (!((WFF)it.next()).isExistsForm()) return false ; 
         else return true ; 
      }
      return false ; 
   }


   /**
     *  @return whether or not this WFF is existential conjunction
     */
   protected boolean isExistsForm() {  
      WFF w = this ;
      // move past all existentials (if any)
      while (w.op == EXISTS) w = (WFF)w.operands.elementAt(1) ;
      if (w.op == FUNCTOR) return true ;     
      else if (w.isAndForm()) return true ; 
      else return false ; 
   }

///////////////////////////////////////////////////////////////////////////  10.1.2011

   /**
     *  Create the " true => '{..reified..}'" rule.
     */
   public static Rule makeAssertion(WFF w,VarList varlist) { 
      Functor tru = new Functor("true",0,Functor.FCNL) ; 
      Conjunction a = new Conjunction() ; 
      a.add(tru) ; 
      Conjunction b = new Conjunction() ; 
      b.add(w.preify(AFT,varlist)) ;  // no vars yet
      Disjunction d = new Disjunction() ; 
      d.add(b) ; 
      Rule rule = new Rule(a,d) ; 
      return rule ; 
   }

   /**
     *  The WFF w will have free variables that 
     *  will be herbrandized, using set.
      */
   protected static void herbrandize(WFF w, HashSet<TypedVariable> set) {
      if (w.op == FUNCTOR) 
         w.operands.set(0, herbrandize((Functor)(w.operands.elementAt(0)), set)) ;
      else if (w.op == IF) {
         herbrandize((WFF)(w.operands.elementAt(0)), set) ;  
         herbrandize((WFF)(w.operands.elementAt(1)), set) ;  
      }
      else if (w.op == AND || w.op == OR) 
         for (int i = 0 ; i < w.operands.size() ; i++) 
            herbrandize((WFF)w.operands.elementAt(i),set) ; 
      else if (w.op == EXISTS) 
            herbrandize((WFF)w.operands.elementAt(1),set) ; 
      else if (w.op == NOT) 
            herbrandize((WFF)w.operands.elementAt(0),set) ; 
   }

   /**
     *  Herbrandize the variables in a term 
     *  which belong to the set.  Variable X 
     *  becomes constant term @X.
     */
   protected static Term herbrandize(Term t, HashSet<TypedVariable> set) { 
      if (t instanceof Variable) {
         if (containsName(((Variable)t).name,set))
            return new Functor("@"+ ((Variable)t).name+hCount,0,Functor.FCNL) ;  // @X
         else return t ; 
      }
      else {
         Functor f = new Functor(((Functor)t).name, ((Functor)t).arity,Functor.FCNL) ;
         f.type = ((Functor)t).type ; 
         for (int i = 0 ; i < f.arity ; i++) {
            f.args[i] = herbrandize(((Functor)t).args[i],set) ; 
         }
         return f ;
      }
   }

   /**
     *  Does the HashSet of TypedVariables have a 
     *  variable with name ...?
      */
   protected static boolean containsName(String s, HashSet<TypedVariable> set) {  
      Iterator it = set.iterator() ;
      while (it.hasNext()) 
         if (s.equals(((TypedVariable)it.next()).name)) return true;
      return false ; 
   }


   protected void pAft(Theory domain,Theory closure,Theory thy,Theory consistency,VarList varlist,TreeSet<String> guards,String method) { 
      if (this.op == FUNCTOR) {
         domainClosure((Functor)this.operands.elementAt(0),domain,closure) ; 
      }
      else if (this.op == AND) {   
         Conjunction ant = new Conjunction() ; 
         ant.add(this.preify(AFT,varlist)) ; 
         Conjunction b = new Conjunction() ; 
         Iterator it = this.operands.iterator() ; 
         while (it.hasNext())   
            b.add(((WFF)it.next()).preify(AFT,varlist)) ; 
         Disjunction d = new Disjunction() ; 
         d.add(b) ;  
         Rule rule = new Rule(ant,d) ; 
         thy.add(rule) ; 
         it = this.operands.iterator() ; 
         while (it.hasNext())   
            ( (WFF)it.next() ).pAft(domain,closure,thy,consistency,(VarList)varlist.clone(),guards,method); //N.B. copy
      }
      else if (this.op == OR) {   
         Conjunction ant = new Conjunction() ; 
         ant.add(this.preify(AFT,varlist)) ; 
         Iterator it = this.operands.iterator() ; 
         Disjunction d = new Disjunction() ; 
         while (it.hasNext())  {
            Conjunction c = new Conjunction() ;
            c.add(((WFF)it.next()).preify(AFT,varlist)) ; 
            d.add(c) ;
         }
         Rule rule = new Rule(ant,d) ; 
         thy.add(rule) ; 
         it = this.operands.iterator() ; 
         while (it.hasNext())   
            ( (WFF)it.next() ).pAft(domain,closure,thy,consistency,(VarList)varlist.clone(),guards,method); 
      }
      else if (this.op == NOT) {   
         Conjunction ant = new Conjunction() ; 
         ant.add(this.preify(AFT,varlist)) ; 
         WFF w = (WFF)this.operands.elementAt(0) ; 
         Disjunction d = new Disjunction() ; 
         Conjunction c = new Conjunction() ;
         c.add(w.nreify(AFT,varlist)) ; 
         d.add(c) ;
         Rule rule = new Rule(ant,d) ; 
         thy.add(rule) ; 
         w.nAft(domain,closure,thy,consistency,(VarList)varlist.clone(),guards,method) ; 
      }
      else if (this.op == IF) {   
        // w1 => w2
        WFF w1 = (WFF)this.operands.elementAt(0) ;
        WFF w2 = (WFF)this.operands.elementAt(1) ;
        Theory tmp_thy = new Theory(),
               tmp_consistency = (Theory)(consistency.clone()) ,
               tmp_domain =(Theory)(domain.clone()),
               tmp_closure = (Theory)(closure.clone()) ; 
        boolean fore_try = false ;
        if (method.contains("fore")) fore_try = true ;  
        if (fore_try) {
           // attempt fore translation of w1
           try { // try to fore-translate antecedent
              w1.pFore(tmp_domain,tmp_closure,tmp_thy,tmp_consistency,(VarList)varlist.clone(),guards,method) ; 
           } 
           catch(ForeException fex) {  fore_try = false ;  }
        }
        if (fore_try) { // Fore translation  was ok
            Conjunction ant = new Conjunction() ; 
            ant.add(this.preify(AFT,varlist)) ; 
            ant.add(w1.preify(FORE,varlist)) ;  
            Disjunction d = new Disjunction() ; 
            Conjunction c = new Conjunction() ; 
            c.add(w2.preify(AFT,varlist)) ; 
            d.add(c) ; 
            Rule rule = new Rule(ant,d) ; 
            thy.add(rule) ; 
            // update from fore-translation
            thy.addAll(tmp_thy) ;   
            consistency.clear() ; // and transfer 
            consistency.addAll(tmp_consistency) ; 
            domain.clear() ; // and transfer          
            domain.addAll(tmp_domain) ;
            closure.clear() ; // and transfer
            closure.addAll(tmp_closure) ; 
            // w2 last, below
        }
        else {   // not Fore translation 
            Conjunction ant = new Conjunction() ; 
            ant.add(this.preify(AFT,varlist)) ; 
            Disjunction d = new Disjunction() ; 
            Conjunction c1 = new Conjunction() ; 
            Conjunction c2 = new Conjunction() ; 
            c1.add(w1.nreify(AFT,varlist)) ; 
            c2.add(w2.preify(AFT,varlist)) ; 
            d.add(c1) ; d.add(c2) ; 
            Rule rule = new Rule(ant,d) ; 
            thy.add(rule) ; 
            w1.nAft(domain,closure,thy,consistency,(VarList)varlist.clone(),guards,method) ; 
         }
         // and finally
         w2.pAft(domain,closure,thy,consistency,(VarList)varlist.clone(),guards,method) ;  
      }
      else if (this.op == IFF) {  
         // >(a<=>b)=> >(a=>b), >(b=>a)
         WFF w1 = (WFF)this.operands.elementAt(0) ;
         WFF w2 = (WFF)this.operands.elementAt(1) ;
         WFF w3 = new WFF(IF) ; 
         w3.operands.add(w1) ; w3.operands.add(w2) ;
         WFF w4 = new WFF(IF) ;
         w4.operands.add(w2) ; w4.operands.add(w1) ; 
         Conjunction ant = new Conjunction() ; 
         ant.add(this.preify(AFT,varlist)) ; 
         Disjunction d = new Disjunction() ; 
         Conjunction c = new Conjunction() ; 
         c.add(w3.preify(AFT,varlist)) ;  
         c.add(w4.preify(AFT,varlist)) ;  
         d.add(c) ; 
         Rule rule = new Rule(ant,d) ; 
         thy.add(rule) ; 
         w3.pAft(domain,closure,thy,consistency,(VarList)varlist.clone(),guards,method) ; 
         w4.pAft(domain,closure,thy,consistency,(VarList)varlist.clone(),guards,method) ; 
      }
      else if (this.op == ALL) {  
         // '+{![X:a,Y:b..]: W}'(..varlist..), a(X), b(Y) .. => '+{W}'(..varlist..,X,Y..).
         Conjunction ant = new Conjunction() ; 
         ant.add(this.preify(AFT,varlist)) ;
         Vector tvars = (Vector)this.operands.elementAt(0) ;
         for(int i=0 ; i < tvars.size() ; i++) { 
            TypedVariable v = (TypedVariable)tvars.elementAt(i) ; 
            guards.add(v.typename) ; 
            Functor e = new Functor(v.typename,1,Functor.FCNL) ;
            //e.args[0]= v.clone() ;
            e.args[0]= v.unType() ;  // new Variable with same name
            ant.add(e) ; 
         }
         // NOW include the new universal variables in varlist
         Iterator newvars = ((Vector)this.operands.elementAt(0)).iterator() ; 
         while (newvars.hasNext()) varlist.add(((TypedVariable)newvars.next()).clone()) ; 
         Conjunction b = new Conjunction() ; 
         WFF w = (WFF)(this.operands.elementAt(1)) ;
         //Functor f = new Functor(w.preify(AFT,varlist),0,Functor.FCNL) ; 
         b.add(w.preify(AFT,varlist)) ; 
         Disjunction d = new Disjunction() ;
         d.add(b) ;  
         thy.add(new Rule(ant,d)) ; 
         w.pAft(domain,closure,thy,consistency,(VarList)varlist.clone(),guards,method) ; 
      }
      else if (this.op == EXISTS) {  
         // '+{?[X:a,Y:b..]: W}'(..varlist..) => a(X), b(Y).., '+{W}'(..varlist..,X,Y..).
         Conjunction ant = new Conjunction() ; 
         ant.add(this.preify(AFT,varlist)) ;
         Conjunction b = new Conjunction() ; 
         Vector tvars = (Vector)this.operands.elementAt(0) ;
         for(int i=0 ; i < ((Vector)this.operands.elementAt(0)).size() ; i++) { 
            TypedVariable v = (TypedVariable)tvars.elementAt(i) ; 
            guards.add(v.typename) ; 
            Functor e = new Functor(v.typename,1,Functor.FCNL) ;
            //e.args[0]= v.clone() ; 
            e.args[0]= v.unType() ;   // new Variable with same name
            b.add(e) ; 
         }
         // NOW include the new variables in varlist
         Iterator newvars = ((Vector)this.operands.elementAt(0)).iterator() ; 
         while (newvars.hasNext()) varlist.add(((TypedVariable)newvars.next()).clone()) ; 
         WFF w = (WFF)(this.operands.elementAt(1)) ; 
         b.add(w.preify(AFT,varlist)) ; 
         Disjunction d = new Disjunction() ;
         d.add(b) ;
         thy.add(new Rule(ant,d)) ; 
         w.pAft(domain,closure,thy,consistency,(VarList)varlist.clone(),guards,method) ; 
      }
   } // end pAft

   protected void nAft(Theory domain,Theory closure,Theory thy,Theory consistency,VarList varlist,TreeSet<String> guards,String method) { 
      if (this.op == FUNCTOR) {
         // negation of a predicate in the aft direction
         consistency( (Functor)this.operands.elementAt(0),consistency) ; 
         domainClosure((Functor)this.operands.elementAt(0),domain,closure) ; 
      }
      else if (this.op == AND) {   
         Conjunction ant = new Conjunction() ; 
         ant.add(this.nreify(AFT,varlist)) ; 
         Conjunction b = new Conjunction() ; 
         Iterator it = this.operands.iterator() ; 
         Disjunction d = new Disjunction() ; 
         while (it.hasNext())  {
            Conjunction c = new Conjunction() ;
            c.add(((WFF)it.next()).nreify(AFT,varlist)) ; 
            d.add(c) ;
         }
         Rule rule = new Rule(ant,d) ; 
         thy.add(rule) ; 
         it = this.operands.iterator() ; 
         while (it.hasNext())   
            ( (WFF)it.next() ).nAft(domain,closure,thy,consistency,(VarList)varlist.clone(),guards,method); 
      }
      else if (this.op == OR) {   
         Conjunction ant = new Conjunction() ; 
         ant.add(this.nreify(AFT,varlist)) ; 
         Conjunction b = new Conjunction() ; 
         Iterator it = this.operands.iterator() ; 
         while (it.hasNext())   
            b.add(((WFF)it.next()).nreify(AFT,varlist)) ; 
         Disjunction d = new Disjunction() ; 
         d.add(b) ;  
         Rule rule = new Rule(ant,d) ; 
         thy.add(rule) ; 
         it = this.operands.iterator() ; 
         while (it.hasNext())   
            ( (WFF)it.next() ).nAft(domain,closure,thy,consistency,(VarList)varlist.clone(),guards,method); 
      }
      else if (this.op == NOT) {   
         Conjunction ant = new Conjunction() ; 
         ant.add(this.nreify(AFT,varlist)) ; 
         WFF w = (WFF)this.operands.elementAt(0) ; 
         Disjunction d = new Disjunction() ; 
         Conjunction c = new Conjunction() ;
         c.add(w.preify(AFT,varlist)) ; 
         d.add(c) ;
         Rule rule = new Rule(ant,d) ; 
         thy.add(rule) ; 
         w.pAft(domain,closure,thy,consistency,(VarList)varlist.clone(),guards,method) ; 
      }
      else if (this.op == IF) {   
         Conjunction ant = new Conjunction() ; 
         ant.add(this.nreify(AFT,varlist)) ; 
         Disjunction d = new Disjunction() ; 
         Conjunction c = new Conjunction() ; 
         WFF w1 = (WFF)this.operands.elementAt(0) ;
         WFF w2 = (WFF)this.operands.elementAt(1) ;
         c.add(w1.preify(AFT,varlist)) ; c.add(w2.nreify(AFT,varlist)) ; 
         d.add(c) ; 
         Rule rule = new Rule(ant,d) ; 
         thy.add(rule) ; 
         w1.pAft(domain,closure,thy,consistency,(VarList)varlist.clone(),guards,method) ; 
         w2.nAft(domain,closure,thy,consistency,(VarList)varlist.clone(),guards,method) ; 
      }
      else if (this.op == IFF) {   
         // >-(a<=>b) => >-(a=>b) | >-(b=>a)
         WFF w1 = (WFF)this.operands.elementAt(0) ;
         WFF w2 = (WFF)this.operands.elementAt(1) ;
         WFF w3 = new WFF(IF) ;
         w3.operands.add(w1) ; w3.operands.add(w2) ; 
         WFF w4 = new WFF(IF) ;
         w4.operands.add(w2) ; w4.operands.add(w1) ; 
         Conjunction ant = new Conjunction() ; 
         ant.add(this.nreify(AFT,varlist)) ; 
         Disjunction d = new Disjunction() ; 
         Conjunction c1 = new Conjunction() ; 
         Conjunction c2 = new Conjunction() ; 
         c1.add(w3.nreify(AFT,varlist)) ;  
         c2.add(w4.nreify(AFT,varlist)) ;  
         d.add(c1) ; d.add(c2) ; 
         thy.add( new Rule(ant,d)) ; 
         w3.nAft(domain,closure,thy,consistency,(VarList)varlist.clone(),guards,method) ; 
         w4.nAft(domain,closure,thy,consistency,(VarList)varlist.clone(),guards,method) ; 
      }
      else if (this.op == ALL) {  
         //    ___this____
         // '-{![X,Y..]: W}'(..varlist..) => '+{?[X,Y..]: ~W}'(..varlist). 
         Conjunction ant = new Conjunction() ; 
         ant.add(this.nreify(AFT,varlist)) ;
         WFF neg = new WFF(NOT) ; 
         neg.operands.add(this.operands.elementAt(1)) ;
         WFF w = new WFF(EXISTS) ; 
         w.operands.add(this.operands.elementAt(0)) ;
         w.operands.add(neg) ; 
         Conjunction b = new Conjunction() ; 
         b.add(w.preify(AFT,varlist)) ; 
         Disjunction d = new Disjunction() ;
         d.add(b) ;  
         thy.add(new Rule(ant,d)) ; 
         w.pAft(domain,closure,thy,consistency,(VarList)varlist.clone(),guards,method) ; 
      } 
      else if (this.op == EXISTS) {  
         // '-{?[X,Y..]: W}'(..varlist..) => '+{![X,Y..]: ~W}'(..varlist). 
         Conjunction ant = new Conjunction() ; 
         ant.add(this.nreify(AFT,varlist)) ;
         WFF neg = new WFF(NOT) ; 
         neg.operands.add(this.operands.elementAt(1)) ;
         WFF w = new WFF(ALL) ; 
         w.operands.add(this.operands.elementAt(0)) ;
         w.operands.add(neg) ; 
         Conjunction b = new Conjunction() ; 
         b.add(w.preify(AFT,varlist)) ; 
         Disjunction d = new Disjunction() ;
         d.add(b) ;  
         thy.add(new Rule(ant,d)) ; 
         w.pAft(domain,closure,thy,consistency,(VarList)varlist.clone(),guards,method) ; 
      } 
   } // end nAft

   /**
     *  positive Fore translations
     *  @throws ForeException
      */
   protected void pFore(Theory domain,Theory closure,Theory thy,Theory consistency,VarList varlist,TreeSet<String> guards,String method) { 
      if (this.op == FUNCTOR) {  // capture free variables in antecedent
         // p(X), dom(Y), dom(Z) => <{p(X)} (X,Y,Z)  
         // find variables in varlist that are not in the functor
         Functor f = (Functor)this.operands.elementAt(0) ; 
                 //HashMap<String,Integer> m = new HashMap<String,Integer>() ; 
                 // f.findVariables(m) ;    
                 //Set vars = m.keySet() ;    CologAnalyzer line 191
         Set vars = CologAnalyzer.findVars(f) ;
         VarList freevars = new VarList() ;  
         Iterator it = varlist.iterator() ; 
         while (it.hasNext()) {
            TypedVariable v = (TypedVariable)it.next() ;
            if (!vars.contains(v)) freevars.add(v.clone()) ; 
         } 
         Disjunction d = new Disjunction() ; 
         Conjunction c = new Conjunction() ; 
         c.add(this.preify(FORE,varlist)) ; 
         d.add(c) ; 
         Conjunction ant = new Conjunction() ; 
         ant.add(f) ; 
         Iterator q = freevars.iterator() ;
         while (q.hasNext()) {  
            TypedVariable v = (TypedVariable)q.next() ;
            Functor guard = new Functor(v.typename,1,Functor.FCNL) ;
            guard.args[0] = v.unType() ; 
            ant.add(guard) ; 
         }         
         thy.add(new Rule(ant,d)) ; 
         domainClosure((Functor)this.operands.elementAt(0),domain,closure) ; 
      }
      else if (this.op == AND) {   
         Conjunction ant = new Conjunction() ; 
         Iterator it = this.operands.iterator() ; 
         while (it.hasNext())   
            ant.add(((WFF)it.next()).preify(FORE,varlist)) ; 
         Disjunction d = new Disjunction() ; 
         Conjunction cons = new Conjunction() ; 
         cons.add(this.preify(FORE,varlist)) ; 
         d.add(cons) ;  
         Rule rule = new Rule(ant,d) ; 
         thy.add(rule) ; 
         it = this.operands.iterator() ; 
         while (it.hasNext())   
            ( (WFF)it.next() ).pFore(domain,closure,thy,consistency,(VarList)varlist.clone(),guards,method); //N.B. copy
      }
      else if (this.op == OR) {   
         Disjunction d = new Disjunction() ; 
         Conjunction conj = new Conjunction() ; 
         conj.add(this.preify(FORE,varlist)) ; 
         d.add(conj) ;
         Iterator it = this.operands.iterator() ; 
         while (it.hasNext())  {
            Conjunction ant = new Conjunction() ;
            ant.add(((WFF)it.next()).preify(FORE,varlist)) ; 
         thy.add(new Rule(ant,d)) ; 
         }
         it = this.operands.iterator() ; 
         while (it.hasNext())   
            ( (WFF)it.next() ).pFore(domain,closure,thy,consistency,(VarList)varlist.clone(),guards,method); 
      }
      else if (this.op == NOT) {   
         WFF w = (WFF)this.operands.elementAt(0) ; 
         Conjunction ant = new Conjunction() ;
         ant.add(w.nreify(FORE,varlist)) ; 
         Disjunction d = new Disjunction() ; 
         Conjunction conj = new Conjunction() ; 
         conj.add(this.preify(FORE,varlist)) ; 
         d.add(conj) ;
         Rule rule = new Rule(ant,d) ; 
         thy.add(rule) ; 
         w.nFore(domain,closure,thy,consistency,(VarList)varlist.clone(),guards,method) ; 
      }   
      else if (this.op == IF) {   
         WFF w1 = (WFF)this.operands.elementAt(0) ;
         WFF w2 = (WFF)this.operands.elementAt(1) ;
         // w1 => w2
         // <-w1 => <(w1 => w2)
         Conjunction conj = new Conjunction() ; 
         conj.add(this.preify(FORE,varlist)) ; 
         Disjunction d = new Disjunction() ; 
         d.add(conj) ; 
         // <w2 => <(w1 => w2)
         Conjunction ant1 = new Conjunction() ; 
         ant1.add(w1.nreify(FORE,varlist)) ;
         thy.add(new Rule(ant1,d)) ;
         Conjunction ant2 = new Conjunction() ; 
         ant2.add(w2.preify(FORE,varlist)) ;
         thy.add(new Rule(ant2,d)) ;
         w1.nFore(domain,closure,thy,consistency,(VarList)varlist.clone(),guards,method) ; 
         w2.pFore(domain,closure,thy,consistency,(VarList)varlist.clone(),guards,method) ; 
      } 
      else if (this.op == IFF) {   
         WFF w1 = (WFF)this.operands.elementAt(0) ;
         WFF w2 = (WFF)this.operands.elementAt(1) ;
         //  <(w1=>w2), <(w2=>w1) => <(w1<=>w2)
         Conjunction conj = new Conjunction() ; 
         conj.add(this.preify(FORE,varlist)) ;
         Disjunction d = new Disjunction() ;
         d.add(conj) ; 
         Conjunction ant = new Conjunction() ; 
         WFF A = new WFF(IF) ; 
         A.operands.add(w1) ;          
         A.operands.add(w2) ;          
         WFF B = new WFF(IF) ; 
         B.operands.add(w2) ;          
         B.operands.add(w1) ;          
         ant.add(A.preify(FORE,varlist)) ; 
         ant.add(B.preify(FORE,varlist)) ; 
         thy.add(new Rule(ant,d)) ; 
         A.pFore(domain,closure,thy,consistency,(VarList)varlist.clone(),guards,method) ; 
         B.pFore(domain,closure,thy,consistency,(VarList)varlist.clone(),guards,method) ; 
      }
      else if (this.op == ALL) {  
         // EXCEPTIONAL CASE
         throw new ForeException(this) ; 
         /////////////////////////////////
      } 
      else if (this.op == EXISTS) {  
         // a(X),'<{w}'(X) => '<{?[X:a]:w'
         // preify this BEFORE adding variables to varlist
         Conjunction b = new Conjunction() ; 
         b.add(this.preify(FORE,varlist)) ;
         Disjunction d = new Disjunction() ;
         d.add(b) ;
         // now add variables
         WFF w = (WFF)this.operands.elementAt(1) ; 
         Conjunction ant = new Conjunction() ; 
         Vector tvars = (Vector)this.operands.elementAt(0) ;
         for(int i=0 ; i < tvars.size() ; i++) { 
            TypedVariable v = (TypedVariable)tvars.elementAt(i) ; 
            varlist.add(v) ; // forgot this previously
            guards.add(v.typename) ; 
            Functor e = new Functor(v.typename,1,Functor.FCNL) ;
            e.args[0]= v.unType() ;  // new Variable with same name
            ant.add(e) ; 
         }
         ant.add(w.preify(FORE,varlist)) ;
         thy.add(new Rule(ant,d)) ; 
         w.pFore(domain,closure,thy,consistency,(VarList)varlist.clone(),guards,method) ; 
      }
   } // end pFore

   /**
     *  negative Fore translations
     *  @throws ForeException
      */
   protected void nFore(Theory domain,Theory closure,Theory thy,Theory consistency,VarList varlist,TreeSet<String> guards,String method) { 
      if (this.op == FUNCTOR) {  // capture free variables in antecedent
         // not_p(X), dom(Y), dom(Z) => <-{p(X)} (X,Y,Z)  
         // find variables in varlist that are not in the functor
         Functor f = (Functor)this.operands.elementAt(0) ; 
                  //HashMap<String,Integer> m = new HashMap<String,Integer>() ; 
                  //f.findVariables(m) ;  
                  //Set vars = m.keySet() ; 
         Set vars = CologAnalyzer.findVars(f) ;
         VarList freevars = new VarList() ;  
         Iterator it = varlist.iterator() ; 
         while (it.hasNext()) {
            TypedVariable v = (TypedVariable)it.next() ;
            if (!vars.contains(v)) freevars.add(v.clone()) ; 
         } 
         Disjunction d = new Disjunction() ; 
         Conjunction c = new Conjunction() ; 
         c.add(this.nreify(FORE,varlist)) ; 
         d.add(c) ; 
         Conjunction ant = new Conjunction() ; 
         ant.add(explicitly_negate(f)) ;   ///// N.B. /////    not_p => ...
         Iterator q = freevars.iterator() ;
         while (q.hasNext()) {  
            TypedVariable v = (TypedVariable)q.next() ;
            Functor guard = new Functor(v.typename,1,Functor.FCNL) ;
            guard.args[0] = v.unType() ; 
            ant.add(guard) ; 
         }         
         thy.add(new Rule(ant,d)) ; 
         domainClosure((Functor)this.operands.elementAt(0),domain,closure) ; 
      }   
      else if (this.op == AND) {   
         Conjunction cons = new Conjunction() ; 
         cons.add(this.nreify(FORE,varlist)) ; 
         Disjunction d = new Disjunction() ; 
         d.add(cons) ;  
         Iterator it = this.operands.iterator() ; 
         while (it.hasNext()) {
            Conjunction ant = new Conjunction() ; 
            ant.add(((WFF)it.next()).nreify(FORE,varlist)) ;
            thy.add(new Rule(ant,d)) ; 
         }
         it = this.operands.iterator() ; 
         while (it.hasNext())   
            ( (WFF)it.next() ).nFore(domain,closure,thy,consistency,(VarList)varlist.clone(),guards,method); //N.B. copy
      }
      else if (this.op == OR) {   
         Disjunction d = new Disjunction() ; 
         Conjunction conj = new Conjunction() ; 
         conj.add(this.nreify(FORE,varlist)) ; 
         d.add(conj) ;
         Iterator it = this.operands.iterator() ; 
         Conjunction ant = new Conjunction() ;
         while (it.hasNext())  {
            ant.add(((WFF)it.next()).nreify(FORE,varlist)) ; 
         }
         thy.add(new Rule(ant,d)) ; 
         it = this.operands.iterator() ; 
         while (it.hasNext())   
            ( (WFF)it.next() ).nFore(domain,closure,thy,consistency,(VarList)varlist.clone(),guards,method); 
      }   
      else if (this.op == NOT) {   
         WFF w = (WFF)this.operands.elementAt(0) ; 
         Conjunction ant = new Conjunction() ;
         ant.add(w.preify(FORE,varlist)) ; 
         Disjunction d = new Disjunction() ; 
         Conjunction conj = new Conjunction() ; 
         conj.add(this.nreify(FORE,varlist)) ; 
         d.add(conj) ;
         Rule rule = new Rule(ant,d) ; 
         thy.add(rule) ; 
         w.pFore(domain,closure,thy,consistency,(VarList)varlist.clone(),guards,method) ; 
      }  
      else if (this.op == IF) {   
         WFF w1 = (WFF)this.operands.elementAt(0) ;
         WFF w2 = (WFF)this.operands.elementAt(1) ;
         // <w1,<-w2 => <-(w1 => w2)
         Conjunction ant = new Conjunction() ; 
         ant.add(w1.preify(FORE,varlist)) ;
         ant.add(w2.nreify(FORE,varlist)) ;
         Conjunction conj = new Conjunction() ; 
         conj.add(this.nreify(FORE,varlist)) ; 
         Disjunction d = new Disjunction() ; 
         d.add(conj) ; 
         thy.add(new Rule(ant,d)) ;
         w1.pFore(domain,closure,thy,consistency,(VarList)varlist.clone(),guards,method) ; 
         w2.nFore(domain,closure,thy,consistency,(VarList)varlist.clone(),guards,method) ; 
      } 
      else if (this.op == IFF) {   
         WFF w1 = (WFF)this.operands.elementAt(0) ;
         WFF w2 = (WFF)this.operands.elementAt(1) ;
         //  <-(w1=>w2) => <-(w1<=>w2)
         //  <-(w2=>w1) => <-(w1<=>w2)
         Conjunction conj = new Conjunction() ; 
         conj.add(this.nreify(FORE,varlist)) ;
         Disjunction d = new Disjunction() ;
         d.add(conj) ; 
         Conjunction antA = new Conjunction() ; 
         WFF A = new WFF(IF) ; 
         A.operands.add(w1) ;          
         A.operands.add(w2) ;          
         antA.add(A.nreify(FORE,varlist)) ;
         thy.add(new Rule(antA,d)) ; 
         Conjunction antB = new Conjunction() ; 
         WFF B = new WFF(IF) ; 
         B.operands.add(w2) ;          
         B.operands.add(w1) ;          
         antB.add(B.nreify(FORE,varlist)) ; 
         thy.add(new Rule(antB,d)) ; 
         A.nFore(domain,closure,thy,consistency,(VarList)varlist.clone(),guards,method) ; 
         B.nFore(domain,closure,thy,consistency,(VarList)varlist.clone(),guards,method) ; 
      } 
      else if (this.op == ALL) {  // similar to pFore EXISTS ?
         // a(X),'<{w}'(X) => '<-{![X:a]:w'
         // preify this BEFORE adding variables to varlist
         Conjunction b = new Conjunction() ; 
         b.add(this.nreify(FORE,varlist)) ;  // N.B.
         Disjunction d = new Disjunction() ;
         d.add(b) ;
         // now add variables
         WFF w = (WFF)this.operands.elementAt(1) ; 
         Conjunction ant = new Conjunction() ; 
         Vector tvars = (Vector)this.operands.elementAt(0) ;
         for(int i=0 ; i < tvars.size() ; i++) { 
            TypedVariable v = (TypedVariable)tvars.elementAt(i) ; 
            varlist.add(v) ; // forgot this previously
            guards.add(v.typename) ; 
            Functor e = new Functor(v.typename,1,Functor.FCNL) ;
            e.args[0]= v.unType() ;  // new Variable with same name
            ant.add(e) ; 
         }
         ant.add(w.nreify(FORE,varlist)) ; // N.B.
         thy.add(new Rule(ant,d)) ; 
         w.nFore(domain,closure,thy,consistency,(VarList)varlist.clone(),guards,method) ; // N.B.
      } 
      else if (this.op == EXISTS) {  
         // EXCEPTIONAL CASE
         throw new ForeException(this) ; 
         ///////////////////////////////
      }
   } // end nFore

   // <<171 (\u00AB) >>187(\u00BB) log.not(\u00AC)
   protected Functor preify(String d, VarList varlist) { // d is FORE or AFT
      if (this.op == FUNCTOR && d == AFT) 
         return (Functor)this.operands.elementAt(0) ; 
      // otherwise decorate 
      String s = "\'"+(d==FORE?FORE:AFT)+"{" + this.toString() + "}\'" ;
      Functor f =  new Functor(s,varlist.size(),Functor.FCNL) ; 
      for(int i=0 ; i < varlist.size() ; i++) f.args[i] = ((Variable)varlist.elementAt(i)).clone() ; 
      return f ; 
   }

   protected Functor nreify(String d, VarList varlist) {  // d is FORE or AFT
      if (this.op == FUNCTOR && d == AFT)  
         return explicitly_negate((Functor)this.operands.elementAt(0)) ; 
      // otherwise decortate
      String s = "\'"+(d==FORE?FORE+"~":AFT+"~")+"{" + this.toString() + "}\'" ;
      Functor f =  new Functor(s,varlist.size(),Functor.FCNL) ; 
      for(int i=0 ; i < varlist.size() ; i++) f.args[i] = ((Variable)varlist.elementAt(i)).clone() ;
      return f ; 
   }

   /**
     * @return contrary ~f(-) of a functor f(-).
     * Explicit negation of '='/2 is '!='/2.
     */
   protected Functor explicitly_negate(Functor f) { 
      // String n = "'-" + f.name + "'" ;  // old way
      String n  ; 
      if (f.name.equals("=") && f.arity==2) n = "!=" ; 
      else if (f.name.equals("!=") && f.arity==2) n = "=" ; 
      else n = "~" + f.name  ;  
      Functor g = new Functor(n,f.arity,Functor.FCNL) ; 
      for (int i=0 ; i<f.arity ; i++) g.args[i] = f.args[i] ;
      return g ;
   }

   /**
     *  Add p(-), '~p'(-) => false. ... if not already present.
     */
   protected void consistency(Functor f, Theory thy) { 
      // no consistency rules for true or dom
      if (f.name.equals("true") || f.name.equals(DOM)) return ; 
      // already present?
      for (int i = 0 ; i < thy.size() ; i++) {
         Rule r = (Rule)thy.elementAt(i) ; 
         Functor g = r.antecedent.elementAt(0) ;
         // compare functors (no index yet)
         if (g.name.equals(f.name) && g.args.length == f.args.length) {
            return ;  // already present so do nothing
         }
      } 
      // otherwise add new consistency rule
      Functor nf = explicitly_negate(f) ; 
      Conjunction ant = new Conjunction() ; 
      ant.add(generalize(f)) ; ant.add(generalize(nf)) ; ///
      Disjunction d = new Disjunction() ; 
      Conjunction c = new Conjunction() ; 
      c.add(new Functor("false",0,Functor.FCNL)) ; 
      d.add(c) ; 
      thy.add(new Rule(ant,d)) ; 
   }

   /**
     *  Generate p(X0,X1,...)
     */
   protected Functor generalize(Functor f) { 
      Functor g = new Functor(f.name,f.arity,Functor.FCNL) ; 
      for (int i = 0 ; i < f.arity ; i++) 
         g.args[i] = new Variable("X"+String.valueOf(i)) ; 
      return g ; 
   }

///////////////////////////////////////////////////////////////
/////////
///////// Generate geolog Theory
/////////   
///////////////////////////////////////////////////////////////
   /** 
     *  Generate geolog theory. 
     *  @param  method is "terse+fore","terse+aft","verbose+fore" or "verbose+aft".
     *  Communicate with translator via method parameter.
      */
   public static Theory colog(Fol fol,String method) { 
      hCount = 0 ; // reinitialize in case GUI reloads
      Theory main = new Theory() , 
             consistency = new Theory() ,
             domain = new Theory(),
             closure = new Theory() ;
      TreeSet<String> guards = new TreeSet<String>() ;  
      ///
      Iterator it = fol.theory.iterator() ; 
      int counter = 0 ;
      while (it.hasNext()) { 
         counter++ ;
         Theory temp = new Theory() ; 
         Formula f = (Formula)it.next() ; 
         WFF w = f.wff ; 
         Vector<String> warnings = f.warnings ; 
         ////////// check for warnings //////////////////////////
         w.check(counter,warnings,new VarList(), new VarList()) ;
         ////////////////////////////////////////////////////////
         w.toGeolog(domain,closure,temp,consistency,guards,method) ;  // <<<<< 
         if (method.contains("terse")) { // TERSE
            Point p = WFF.starts(temp) ; 
            // remove first p.x rules
            for (int i = 0 ; i <= p.x-1 ; i++)   
               temp.remove(0) ;
            // save start rule 
            Rule r = temp.elementAt(0) ; 
            // remove start rule
            temp.remove(0) ; 
            // modify start rule
            Conjunction a = new Conjunction() ; 
            if (p.y == 1) a.add(new Functor("true",0,Functor.FCNL)) ; 
            else      // <<< fix 
               for(int i=1 ; i < r.antecedent.size() ; i++) 
                  a.add(r.antecedent.elementAt(i)) ;
            // add modified start rule
            temp.add(0,new Rule(a,r.consequent)) ; 
            ///////////////////////////////////////////////////////////////////
            //  It is possible that some rules has quoties both fore and aft
            // This is not an optimal fix ???
            Theory saveQuotedAntecedents = findQuotedAntecedents(temp) ; // KLUDGE 1
            temp.addAll(saveQuotedAntecedents) ; // KLUDGE 2 see endnote for 12.23.2102
            foreFold(temp) ;     // use folding to rewrite Antecedents of rules
            factor(temp) ;       // remove unneeded factors from antecedents
            domTrim2(temp) ;     // remove unneeded doms from antecedents
            temp.addAll(saveQuotedAntecedents) ; // KLUDGE 2 REDUX retrieve possible missing rules
            aftFold(temp) ;      // use folding to rewrite consequents of rules
            ///////////////////////////////////////////////////////////////////
         }
         // now add temp to main
         Iterator it2 = temp.iterator() ; 
         while (it2.hasNext()) main.add((Rule)it2.next()) ;          
      }  
      // could sort main ?
      // domain + consistency + main + closure
      // consistency axioms at front
      for(int j = consistency.size()-1 ; j >=0 ; j--) // reverse order
         main.add(0,(Rule)consistency.elementAt(j)) ; // at front
      // domain axioms before consistency axioms
      for(int j = domain.size()-1 ; j >=0 ; j--)  // reverse order
         main.add(0,(Rule)domain.elementAt(j)) ;  // at front
      ///// guards => dom
      Iterator grds = guards.iterator() ; 
      while (grds.hasNext()) {
         String type = (String)grds.next() ; 
         //  t(x) => dom(X) for quant guard t
         if (!type.equals(DOM)) main.add(subDom(type)) ; 
      }
      ///// function closures
      Iterator clsrs = closure.iterator() ; 
      while (clsrs.hasNext()) main.add((Rule)clsrs.next()) ; 
      /////
      // return main ;   NO!
      ///////////////////////////////////////////////////////
      // Using main directly, verbose version, causes trouble
      // for more recent colog. However, its toString() is good.
      // so read from the string to get the colog theory, for now.
      // 11.1.2014
      Theory theory = null ; 
      try {
         InputStream in = new ByteArrayInputStream(main.toString().getBytes()) ;
         cologLoader loader = new cologLoader(in) ;    
         if (loader.theory == null || loader.theory.size()==0 || loader.error) 
            theory = null ; 
         else theory =   loader.theory ;
         in.close() ;
      }
      catch (Exception xxx) { System.out.println("Cannot convert FOL to colog.") ; }
      return theory ;
   }


   /*
      Folding can produce repeated factors in an antecedent, e.g.,
      ![X:a]:(p(X) => q(X)).   
      a(X), p(X), a(X) => q(X).
   */

   /**
     * Factor antecedents of all rules of theory.
     *  N.B., need construct new Rule so that Rule data A,C,U,V is accurate.
      */
   static void factor(Theory thy) {
      for (int i = 0 ; i < thy.size(); i++)  {
         Rule r = thy.elementAt(i) ; 
         thy.set(i, new Rule(factor(r.antecedent),r.consequent)) ;
      }
   }

   /**
     *  factor a conjunction
      */
   protected static Conjunction factor(Conjunction c) {
      Conjunction cc = new Conjunction() ; 
      Iterator it = c.iterator() ;
      while (it.hasNext()) {
         Functor f = (Functor)it.next() ; 
         // add if not a repeat
         if (!cc.contains(f)) cc.add(f) ; 
      }
      return cc ;
   }

   /**
     *  a(Z) => dom(Z)  for quantifier guard (type) a/1
      */
   protected static Rule subDom(String s) {
      Variable z = new Variable("Z") ; 
      Functor a = new Functor(s,1,Functor.FCNL) ;
      a.args[0] = z ;
      Functor dom = new Functor(DOM,1,Functor.FCNL) ;
      dom.args[0] = z ;
      Conjunction ant = new Conjunction() ;
      ant.add(a) ;
      Conjunction cons = new Conjunction() ; 
      cons.add(dom) ; 
      Disjunction d = new Disjunction() ;
      d.add(cons) ; 
      return new Rule(ant,d) ; 
   }

   /**
     *  Fold the theory using rewriting:
     *  L = <C1 => 'quoty'(X), C2 => 'quoty'(X), ...>
     *  is used to subtitute Ci for 'quoty'(X) 
     *  in ANTECEDENTS (fore) in all the rules.
      */
   protected static void foreFold(Theory temp) {
      while (true) {  // until no more quoties
         Theory quoties = findQuotedConsequents(temp) ; // removes quoties from temp
         if (quoties == null) break ;  // no more rewrites possible
         // use quoties to rewrite remaining rules
         Functor q = quoties.elementAt(0).consequent.elementAt(0).elementAt(0) ; 
         Rule r = null ; int p = -1 ; int rulenum = -1 ;  int k = 0 ;
         // find rule/positions of rule to rewrite 
         boolean more = true ;  // starts loop
         while (more) {   
            more = false ;   // unless find more
            for (int j = k ; j < temp.size() ; j++) {
               rulenum = j ; 
               r = temp.elementAt(j) ; 
               p = locate(q,r) ;
               if (p > -1) { 
                  more = true ;
                  break ; 
               }
            } 
            if (more) {
               Theory thy = rewrite(r,quoties,p) ; 
               temp.removeElementAt(rulenum) ; 
               temp.addAll(rulenum,thy) ; // replace at rulenum
               k = rulenum + thy.size() ; // the next possible rewrite
            }
            //else break ; // while 
         }
      }  
   }

   // fore rewrite r using quoties, replacing functor at loc
   static Theory rewrite(Rule r,Theory quoties,int p) {
      Theory thy = new Theory() ; 
      for (int i = 0 ; i < quoties.size() ; i++) {
         Conjunction c = (Conjunction) r.antecedent.clone() ; 
         c.removeElementAt(p) ; 
         c.addAll(p,quoties.elementAt(i).antecedent) ;
         thy.add(new Rule(c,r.consequent)) ;
      }
      return thy ; 
   }

/////////////////////////////////////////////////////////////////////////
   /**
     *  AFTER FACTOR, remove any dom(Z) if Z 
     *  occurs in another predicate of antecedent.
      */
   static void domTrim2(Theory thy) { 
      for (int i = 0 ; i < thy.size() ; i++) { 
         Rule r = thy.elementAt(i) ;
         Conjunction c = r.antecedent ;
         if (hasDOM(c)) 
            thy.setElementAt(new Rule(domTrim2(c),r.consequent),i) ;
      }
  }

   /*** Does c have a dom in it? */
   static boolean hasDOM(Conjunction c) { 
      for (int i = 0 ; i < c.size(); i++) 
         if (c.elementAt(i).name.equals(DOM)) return true ; 
      return false ; 
   }

   /*** remove all non-free doms from c */ 
   static Conjunction domTrim2(Conjunction c) { 
      Conjunction trimmed = new Conjunction() ; 
      for(int i = 0 ; i < c.size() ; i++) {
         Functor f = c.elementAt(i) ; 
         if (f.name.equals(DOM)) {
            Variable v = (Variable)f.args[0] ; // f's variable 
            if (!occurs(v,c,i))     // v does NOT occur in another predicate
               trimmed.add(f) ; // necessary universal
            // otherwise f is unneeded
         }
         else trimmed.add(f) ;  // not a dom predicate
      }           
      return trimmed ; 
   }

   /*** Does v occur at a factor other than i? */
   static boolean occurs(Variable v,Conjunction c,int i) {
      for (int j = 0 ; j < c.size() ; j++) 
         if (j != i && occurs(v,c.elementAt(j))) return true ;
      return false ; 
   }

//////////////////////////////////////////////////////////////////////////

   static boolean occurs(Variable x, Term t) {
      if (t.isVariable()) {
         if (((Variable)t).name.equals(x.name)) 
            return true ; 
         else return false ;
      }
      else { // t is a functor
         Functor f = (Functor)t ;
         for (int i = 0 ; i < f.args.length ; i++) 
            if (occurs(x,(Term)f.args[i])) return true ; 
      }
      return false ;
   }

   // at what position in antecedent of r does q occur? -1 = doesn't
   static int locate(Functor q, Rule r) { 
      Conjunction c = r.antecedent ;
      for (int i = 0 ; i < c.size(); i++) {
         //if (c.elementAt(i).index == q.index) // N.B. uses .index
         if (c.elementAt(i).name.equals(q.name)) 
            return i ; 
      }
      return -1 ; // q not found
   }

   static Theory findQuotedConsequents(Theory temp) {
      Theory quoties = new Theory() ;  
      Functor quoty = null ; // will find
      for (int i = 0 ; i < temp.size() ; i++) { 
         Rule r = temp.elementAt(i) ; 
         quoty = quotedConsequent(r) ; 
         if (quoty != null) break ; 
      } // end for
      if (quoty == null) return null ; // no more
      // otherwise find other quoties
      for (int i = 0 ; i < temp.size() ; i++) { 
         Rule r = temp.elementAt(i) ; 
         Functor f = quotedConsequent(r) ; 
         if (f != null && f.name.equals(quoty.name) && f.arity == quoty.arity) // functor equals
            quoties.add(r) ;
      }   
      // extract quoties from temp
      for (int i = 0 ; i < quoties.size() ; i++) { 
         temp.remove(quoties.elementAt(i)) ;   
      }   
      return quoties ;   // will be null if no more quoties
   }

   // returns null or first quoted predicate
   static Functor quotedConsequent(Rule r) { 
         if (r.consequent.size() == 1 && 
                r.consequent.elementAt(0).size() == 1 &&                     
                   ( ((Functor)r.consequent.elementAt(0).elementAt(0)).name.startsWith("'"+AFT) 
                     | ((Functor)r.consequent.elementAt(0).elementAt(0)).name.startsWith("'"+FORE)  ) ) 
            return ((Functor)r.consequent.elementAt(0).elementAt(0)) ;
         else return null ; 
  }

/////////////////////   aft folding /////////////////////
   /**
     * Fold the theory using rewriting:
     *  'quoty'(X) => C    (C a conjunction)
     * is used to subtitute C for 'quoty'(X) 
     * in consequents (aft) in all the rules.
      */
   protected static void aftFold(Theory temp) {
      Theory quoties = findQuotedAntecedents(temp) ;  // quoty => a,b,c3
      // use quoties to rewrite all consequences
      for (int i = 0 ; i < temp.size() ; i++)  // each rule
         for (int j = 0 ; j < quoties.size() ; j ++) 
            // rewrite ith rule using each jth quotie
            temp.setElementAt(rewrite(temp.elementAt(i),quoties.elementAt(j)),i) ;
   }

/* e.g.
 true => dom(X), '{((p(X) & q(X)) | ~r(X))}'(X).
 '{((p(X) & q(X)) | ~r(X))}'(X) => '{(p(X) & q(X))}'(X) | '{~r(X)}'(X).
 '{(p(X) & q(X))}'(X) => p(X), q(X).
 '{~r(X)}'(X) => ~r(X).
*/

   // aft rewrite of rule
   static Rule rewrite(Rule r, Rule quoty) { 
      Disjunction d = new Disjunction() ; 
      for (int i = 0 ; i < r.consequent.size() ; i++) 
         d.add(rewrite(r.consequent.elementAt(i),quoty)) ;
      return new Rule(r.antecedent,d) ; 
   } 

   // aft rewrite of rule
   static Conjunction rewrite(Conjunction c, Rule quoty) { 
      Conjunction cc = new Conjunction() ;  
      for (int i = 0 ; i < c.size() ; i++) { 
         Functor f = c.elementAt(i) ; 
         //if (f.index == quoty.antecedent.elementAt(0).index)  // N.B. uses .index
         if (f.name.equals(quoty.antecedent.elementAt(0).name))  
            cc.addAll(quoty.consequent.elementAt(0)) ; 
         else cc.add(f) ; 
      }
      return cc ; 
   }

   // quoty => Conjunction
   static Theory findQuotedAntecedents(Theory temp) {
      Theory quoties = new Theory() ;
      for (int i = 0 ; i < temp.size() ; i++) {
         Rule r = temp.elementAt(i) ; 
         if (r.antecedent.size() == 1 &&
               ( r.antecedent.elementAt(0).name.startsWith("'"+AFT) | 
                       r.antecedent.elementAt(0).name.startsWith("'"+FORE) ) &&
                  r.consequent.size() == 1 ) { // conjunctive
            quoties.add(r) ; 
         }
      } 
      Iterator it = quoties.iterator() ; 
      while (it.hasNext()) temp.remove(it.next()) ; // object removal by ==
      return quoties ; 
   }

//////////////////////////////////////////////////////////////////////////////////////////

   /** 
     *     0      ...   (kth rule,antecedent.size())
     *   true => ... => --
     */
   protected static Point starts(Theory thy) { 
      for(int k = 0 ; k < thy.size() ; k++) 
         if (thy.elementAt(k).antecedent.size() > 1 ||  
              thy.elementAt(k).consequent.size() > 1  ||
               thy.elementAt(k).consequent.elementAt(0).size() > 1)
            return new Point(k,thy.elementAt(k).antecedent.size())  ; 
      return new Point(thy.size()-1,thy.elementAt(thy.size()-1).antecedent.size()) ; 
   }

   /**
     *  Throw ForeException when translating formal implications.
      */
   protected class ForeException extends RuntimeException {
      WFF wff ; 
      protected ForeException(WFF w) { this.wff = w ; }
   }


 
///////////////////////////////////////////////////////////////////////////////////////////////
   /**
     * Check WFF for correct form.
     * A:Repeat variable is misleading.  
     *   (?[X:a]:p(X) & ?[X:b]:p(X)).
     * B:Unscoped variable is misleading. 
     *   ![X]:(p(X) => q(X,Y)).
      */
   public void check(int counter,Vector<String> warnings,VarList allvars,VarList scopevars) { 
      if (this.op==FUNCTOR)  
         check(counter,warnings,(Functor)this.operands.elementAt(0),scopevars) ; 
      else if (this.op==AND || this.op == OR || this.op==NOT || this.op==IF || this.op==IFF) 
         for(int i = 0 ; i < this.operands.size() ; i++)
            ((WFF)this.operands.elementAt(i)).check(counter,warnings,allvars,scopevars) ; 
      else if (this.op==ALL || this.op==EXISTS) {
         Vector vars = (Vector)this.operands.elementAt(0) ; 
         VarList newscopevars = new VarList() ; 
         // copy scopevars into newscopevars
         Iterator scopeit = scopevars.iterator() ; 
         while (scopeit.hasNext()) newscopevars.add((TypedVariable)scopeit.next()) ;
         Iterator it = vars.iterator() ; 
         while (it.hasNext()) {
            TypedVariable v = (TypedVariable)it.next() ; 
            newscopevars.add(v) ; 
            if (appears(v,allvars)) { // v already used  A
               String warn = " WARNING: variable " + v.name + " appears in more than one quantifier" ;
               System.out.println("AXIOM#"+counter+warn) ; 
               warnings.add("  //"+warn) ; 
            }
            allvars.add(v) ;
         }
         ((WFF)this.operands.elementAt(1)).check(counter,warnings,allvars,newscopevars) ;
      }
   }

   /**
     *  Does variable appear by name in list?
     *  Note that VarList.contains(-) would not work.
      */
   protected boolean appears(Variable v, VarList varlist) { 
      //System.out.println(varlist) ; 
      Iterator it = varlist.iterator() ; 
      while (it.hasNext()) {
         TypedVariable tvar = (TypedVariable)it.next() ;
         if (v.name.equals(tvar.name)) return true ;
      }
      return false ;
   }

   // B
   protected void check(int counter,Vector<String> warnings,Term t, VarList varlist) {
      if (t.isVariable()) {
         if (!appears((Variable)t, varlist) ) {
            String warn = " WARNING: variable "+ t + " has free occurrence" ;
            System.out.println("AXIOM#"+counter+warn) ; 
            warnings.add("  //"+warn) ; 
         }
      }
      else // t is Functor
         for (int i = 0 ; i < ((Functor)t).args.length ; i++) 
            check(counter,warnings,((Functor)t).args[i],varlist) ;
   }


///////////////////////////////////////////////////////////////////////////////

   // stub test factoring
   public static void main(String[] args) { 
   }


////////////////////////////////////////
// Domain axioms and function closures
///////////////////////////////////////////////////////////////////////////////
   /**
     *  Assert domain and closure rules: (e.g.)<br>
     *    true => dom(a).<br
     *    dom(X), dom(Y) =&lt; dom(f(X,Y)). <br>
     *  etc.
     */
   protected void domainClosure(Functor f, Theory domain, Theory closure) { 
      for (int i = 0 ; i < f.arity ; i++) {
         Term t = f.args[i] ; 
         if (!t.isVariable()) { 
            assertDomain((Functor)t, domain, closure) ; 
            domainClosure((Functor)t, domain, closure) ;  
         }
      }
   }

   /**
     *  Domain closures.
     */
   protected void assertDomain(Functor f, Theory domain, Theory closure) { 
      if (f.arity == 0) {
         Functor t = new Functor("true",0,Functor.FCNL) ; 
         Conjunction ant = new Conjunction() ; 
         ant.add(t) ; 
         Disjunction d = new Disjunction() ; 
         Conjunction c = new Conjunction() ; 
         Functor e = new Functor(DOM,1,Functor.FCNL) ; 
         e.args[0] = f ;
         c.add(e) ; 
         d.add(c) ;
         if (!redundant(f,domain)) 
            domain.add(new Rule(ant,d)) ;   
      }
      else { 
         Functor f1 = generalize(f) ; 
         Conjunction ant = new Conjunction() ; 
         for (int i = 0 ; i < f1.arity ; i++) {
            Functor e = new Functor(DOM,1,Functor.FCNL) ; 
            e.args[0] = f1.args[i] ;
            ant.add(e) ; 
         }
         Disjunction d = new Disjunction() ; 
         Conjunction c = new Conjunction() ; 
         Functor g = new Functor(DOM,1,Functor.FCNL) ; 
         g.args[0] = f1 ; 
         c.add(g) ; d.add(c) ; 
         if (!redundant(f,closure) && !f.name.startsWith("$") ) // NOCLOSE$
              closure.add(new Rule(ant,d)) ; 
      }
   }

   /**
     *  Is closure dom(f...) already present?
     */
   protected boolean redundant(Functor f, Theory thy) { 
      boolean found = false ;  
      Iterator it = thy.iterator() ; 
      while (it.hasNext()) { 
         Rule r = (Rule)it.next() ;  
         // ?? check antecedent ??
         Functor q = (Functor)( r.consequent.elementAt(0).elementAt(0));
         if (q.name.equals(DOM) 
              && !q.args[0].isVariable()    // N.B.
               && f.arity == ((Functor)q.args[0]).arity 
                && f.name.equals(((Functor)q.args[0]).name))  { 
            found = true ; 
            break ;
         }
      }   
     return found ; 
   }



} // end WFF
///////////////////////////////////////////////////////////////////////////////

/**
  * List of TypedVariable.
  * These vars all arise from quantifications,
  * either universal or existention.
   */
class VarList extends Vector<TypedVariable>  {}  

//////////////////////////////////////////////////////////////////////////////////


/*
12/23/1012

Terse translation of 
   ((p|~p) => ~(p|~ p)). 
showed a bug.


*1* p, ~p => false.
/*2* true => '>>{((p | ~p) => ~(p | ~p))}'.
*3* '>>{((p | ~p) => ~(p | ~p))}' => '>>~{(p | ~p)}' | '>>{~(p | ~p)}'.
*4* '>>~{(p | ~p)}' => ~p, '>>~{~p}'.
*5* '>>~{~p}' => p.
*6* '>>{~(p | ~p)}' => '>>~{(p | ~p)}'.   AHA! both fore and aft quoted!
*7* '>>~{(p | ~p)}' => ~p, '>>~{~p}'.
*8* '>>~{~p}' => p.


quoted_consequents:  
   2 6 

rewrite antecedents
*3* true => '>>~{(p | ~p)}' | '>>{~(p | ~p)}'.  by 2

quoted_antecedents:  
   4 5 (6 is gone) 7 8     HERE IS THE PROBLEM

rewrite if 6 were available
*3* true => '>>~{(p | ~p)}' | '>>{~(p | ~p)}'.  by 6 IF IT WERE AVAILABLE


*/

