package sam.gui ; 

import sam.lang.antlr.* ;
import sam.lang.* ; 
import sam.infer.* ;
import sam.util.* ;

import java.util.* ;
import java.io.* ; 
import java.net.* ;
import java.awt.* ;
import java.awt.event.* ;
import java.awt.datatransfer.* ; 
import javax.swing.* ;
import javax.swing.event.* ;
import javax.swing.plaf.metal.* ;

/**
  *  Colog GUI management 
  *  @version 5 October 2014 (rev. 1/31/2023 use DefaultListModel<String>)
  *  @author &copy; John Fisher 
   */
public class Colog extends JFrame 
                   implements ActionListener, ChangeListener, ListSelectionListener  { 

   public int cologProcessNumber ;
   private JMenuItem reloadMenuItem ; 
   private JButton reloadButton ;    
   //////////////////////////
   private JToolBar toolbar ; 
   //////////////////////////
   public static String BANNER = CologConstants.BANNER ;     
   public static String NOMATCH = CologConstants.NOMATCH ; // bottom rule "choice"

   protected Theory theory ; // loaded from file or URL
   protected Tree tree ; 
   protected String theoryDescription = "" ;
   protected DefaultListModel<String> theoryListModel ;  
   // to left
   protected JTabbedPane tabbedPane ; // to hold treePane, consoleArea, auxArea //xxxxx
   protected TreePane treePane ;
   protected JTextArea consoleArea ; // to show console messages //xxxxx
   protected JTextArea auxArea ;   // to show proofs  //xxxxx
   protected FOLPane folTool ;     // to manage fol theories and translation
   protected JEditorPane infoPane ; 
   // to right
   protected JList theoryPane ;  
   protected Color shade =  
           //new Color(245,255,255) ; // green-blue/volkswagen
           //new Color(230,255,230) ; 
           //new Color(250,250,250) ;
           //new Color(228,252,236) ;
           new Color(250,255,250) ;   // greenish
   protected JLabel ruleLabel ; 
   protected JButton runButton ;
   protected JMenuItem runMenuItem ;
   protected JButton clearButton ;
   protected JMenuItem clearMenuItem ;
   protected JScrollPane treeScroller ; 
   protected JSplitPane splitPane ;  
   protected JToggleButton complexityButton ; // term complexity heuristic
   protected JSpinner cspinner ; 
   protected JLabel boundsLabel ;
   private int keyAccelerator ; 
   private String fontName ; 
   private boolean toggle ;
   private int treeDepth ; 
   private int treeWidth ;    
   private final Vector<Process> processes ; 
   private final Vector<Boolean> alive ; 
   private static int D = 200 ;
   private static int W = 700 ;
   // GUI window box defaults
      private int x, y, w, h ; 
      private Properties settings ; 
      private String command ; 
   ///////////////////////////////////
   /**
     *  Construct the GUI and event handlers.
     */ 
   public Colog(int cpn) { 
      settings = new Properties() ; 
      try { 
         settings.load(
           new FileInputStream(
            System.getProperty("user.home")+File.separator+"colog.settings" )  ) ; 
         this.x = Integer.parseInt(settings.getProperty("x").trim()) ;
         this.y = Integer.parseInt(settings.getProperty("y").trim()) ; 
         this.w = Integer.parseInt(settings.getProperty("w").trim()) ;
         this.h = Integer.parseInt(settings.getProperty("h").trim()) ; 
         //this.command = settings.getProperty("command").trim() ;
      }
      catch(Exception any_trouble) {x = 300 ; y = 150 ; w = 800 ; h = 600 ;  System.out.println(any_trouble) ; } // defaults 
      command = "java -Xms1g -Xmx1g -jar " + 
              System.getProperty("user.home") + 
              System.getProperty("file.separator") +   
              CologConstants.JARNAME ;  
      /////////////////////////////////
      processes = new Vector<Process>() ; 
      alive =  new Vector<Boolean>() ; 
      //System.out.println(System.getProperty("user.dir")) ;
      //System.out.println(System.getProperty("user.home")) ;
      //System.out.println(System.getProperty("user.name")) ;
      this.cologProcessNumber = cpn ;
      //sharedCount = new SharedCount() ; 
      //System.out.println(System.getProperty("os.name")) ; 
      this.treeDepth = D ;    // default
      this.treeWidth = W ;    // same as tree.width 
      this.setTitle(BANNER + theoryDescription) ; 
      // TREE & TREE PANE 
      tree = new Tree(this.treeDepth,this.treeWidth) ;   
      tree.setLogging(true) ;   // logging true when using GUI
      treePane = new TreePane() ;   // 
      //treePane.setBackground(new Color(200,250,250)) ; 
      treePane.setCologTree(tree) ; 
      treeScroller =   //xxxxx
         new JScrollPane(treePane) ; 
      treePane.scrollPane = treeScroller ;   // for view dragging  //xxxxx
      tabbedPane = new JTabbedPane() ;   //xxxxx
      consoleArea = new JTextArea() ;  //xxxxx 
        consoleArea.setToolTipText("<html><font size=\"+0\"><b>console area</b><br> &nbsp;reports for load errors,&nbsp;&nbsp;   <br>&nbsp;system reports, <<br> &nbsp;and search reports.&nbsp;&nbsp;     ") ;
      consoleArea.setBackground(new Color(240,240,240)) ; 
      auxArea = new JTextArea() ;   //xxxxx  
      auxArea.setToolTipText("<html><font size=\"+0\"><b>aux area</b><br> &nbsp;proof reports,&nbsp;&nbsp;   <br>&nbsp;model reports, <br>&nbsp;theory analysis, <br> &nbsp;LaTeX tree picture code.&nbsp;&nbsp;     ") ;
      auxArea.setBackground(new Color(255,255,204)) ; 
      folTool = new FOLPane(this) ; 
      Class clz = this.getClass() ; 
      URL info = clz.getResource("Info.html") ;
      try { infoPane = new JEditorPane(info) ; } catch(Exception xxx) { System.out.println(xxx) ; }
      infoPane.setBorder(BorderFactory.createMatteBorder(8,8,0,3,shade));
      infoPane.setBackground(shade) ; 
      tabbedPane.addTab("tree", null, treeScroller, "view colog tree") ;  //xxxxx
      tabbedPane.addTab("console", null, new JScrollPane(consoleArea), "view colog console") ; //xxxxx
      tabbedPane.addTab("aux", null, new JScrollPane(auxArea), "view proof, models, theory analysis, LaTeX code") ; //xxxxx
      tabbedPane.addTab("FOL tool",null,folTool,"manage fol inputs") ; 
      tabbedPane.addTab("? ? ?", null, new JScrollPane(infoPane), "view colog info") ; //xxxxx
      try { 
         UIManager.setLookAndFeel(UIManager.getSystemLookAndFeelClassName()); 
         //UIManager.setLookAndFeel(new MetalLookAndFeel()); 
      } catch(Exception x) {}

      if (System.getProperty("os.name").indexOf("Mac") >= 0) {
         fontName = "Monaco" ; //  My favorite on the Mac
         keyAccelerator = InputEvent.META_DOWN_MASK ; 
      }
      else {
         fontName = "sanserif" ;  // do better if you can 
         keyAccelerator = InputEvent.CTRL_DOWN_MASK ; 
      }

      this.theoryListModel = new DefaultListModel<String>() ;
      this.setDefaultCloseOperation(WindowConstants.DO_NOTHING_ON_CLOSE) ; // exitExam()
      this.setSize(w,h) ; 
      this.setLocation(x,y) ; 

      // Create a menubar 
      JMenuBar mb = new JMenuBar() ; 
      this.setJMenuBar(mb)  ; 
      // FILE menu
      JMenu fileMenu = new JMenu("Theory") ; 
      // LOAD colog file
      JMenuItem openMenuItem = new JMenuItem("Open file via dialog") ; 
      openMenuItem.addActionListener(this) ;
      openMenuItem.setActionCommand("open") ;
      openMenuItem.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_O,keyAccelerator)) ;  // InputEvent.META_DOWN_MASK
      fileMenu.add(openMenuItem) ; 
      // LOAD colog from URL
      JMenuItem openURLMenuItem = new JMenuItem("open URL via dialog") ; 
      openURLMenuItem.addActionListener(this) ;
      openURLMenuItem.setActionCommand("openURL") ;
      openURLMenuItem.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_U,keyAccelerator)) ;  // InputEvent.META_DOWN_MASK
      fileMenu.add(openURLMenuItem) ; 
      // reLOAD 
      reloadMenuItem = new JMenuItem("Reload") ; 
      reloadMenuItem.addActionListener(this) ;
      reloadMenuItem.setActionCommand("reload") ;
      reloadMenuItem.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_R,keyAccelerator)) ;  // InputEvent.META_DOWN_MASK
      fileMenu.add(reloadMenuItem) ; 
      JMenuItem newMenuItem = new JMenuItem("New colog process") ; 
      newMenuItem.addActionListener(this) ;
      newMenuItem.setActionCommand("new") ;
      newMenuItem.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_N,keyAccelerator)) ;  // InputEvent.META_DOWN_MASK
      if (cologProcessNumber == 0) {
         fileMenu.addSeparator() ; 
         fileMenu.add(newMenuItem) ; 
      }  // otherwise does not appear
      mb.add(fileMenu) ; 
      // RUN menu
      JMenu runMenu = new JMenu("run") ; 
      // RUN  
      runMenuItem = new JMenuItem("Run =>") ; 
      runMenuItem.addActionListener(this) ;
      runMenuItem.setActionCommand("run") ;
      runMenuItem.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_R,0)) ;  
      //runMenuItem.setEnabled(false) ; 
      runMenu.add(runMenuItem) ; 
      // CLEAR  
      clearMenuItem = new JMenuItem("Clear tree") ; 
      clearMenuItem.addActionListener(this) ;
      clearMenuItem.setActionCommand("clear") ;
      clearMenuItem.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_C,0)) ;  
      runMenu.add(clearMenuItem) ; 
      mb.add(runMenu) ; 

      // OPTIONS menu
      JMenu optionsMenu = new JMenu("options") ; 
      // SHOW TOOLBAR OR NOT
      JCheckBoxMenuItem toolbarTextMenuItem = 
           new JCheckBoxMenuItem("Toolbar",true) ; 
      toolbarTextMenuItem.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_T,keyAccelerator)) ; 
      toolbarTextMenuItem.addActionListener(this) ;  
      optionsMenu.add(toolbarTextMenuItem) ; 
      // RUN BOUNDS
      JMenuItem bounds = new JMenuItem("set depth/width Bounds") ; 
      bounds.addActionListener(this) ; 
      bounds.setActionCommand("bounds") ; 
      bounds.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_B,0)) ;  
      optionsMenu.add(bounds) ; 
      // SCALE menuitem
      JMenu treescaleMenu = new JMenu("tree scale") ; 
      /////
      JMenuItem biggerMenuItem = new JMenuItem("larger +") ; 
      biggerMenuItem.addActionListener(treePane) ;
      biggerMenuItem.addActionListener(this) ;  // for "orient"
      biggerMenuItem.setActionCommand("larger") ;
      biggerMenuItem.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_P,0)) ;  // InputEvent.META_DOWN_MASK
      treescaleMenu.add(biggerMenuItem) ; 
      /////
      JMenuItem smallerMenuItem = new JMenuItem("smaller -") ; 
      smallerMenuItem.addActionListener(treePane) ;
      smallerMenuItem.addActionListener(this) ;  // for "orient"
      smallerMenuItem.setActionCommand("smaller") ;
      smallerMenuItem.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_M,0)) ;  // InputEvent.META_DOWN_MASK
      treescaleMenu.add(smallerMenuItem) ; 
      /////
      JMenuItem fit = new JMenuItem("Scale tree to viewport") ; 
      fit.addActionListener(treePane) ; 
      fit.addActionListener(this) ;
      fit.setActionCommand("fit") ; 
      fit.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_S,0)) ;  
      treescaleMenu.add(fit) ; 
      /////
      JMenuItem original = new JMenuItem("Original scale") ; 
      original.addActionListener(treePane) ; 
      original.addActionListener(this) ;
      original.setActionCommand("original") ; 
      original.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_O,0)) ;  
      treescaleMenu.add(original) ; 
      /////
      /////
      JMenuItem fractionMenuItem = new JMenuItem("50% original") ; 
      fractionMenuItem.addActionListener(treePane) ;
      fractionMenuItem.addActionListener(this) ;
      fractionMenuItem.setActionCommand("fraction") ;
      fractionMenuItem.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_F,0)) ;  // InputEvent.META_DOWN_MASK
      treescaleMenu.add(fractionMenuItem) ; 
      optionsMenu.add(treescaleMenu) ; 
      ///// ZOOM to selected node  -- just moves view, no enlargement
      JMenuItem zoom = new JMenuItem("zoom selected") ; 
      zoom.createToolTip() ; 
      zoom.setToolTipText("<html><font size=\"+0\">move view to see selected tree fact") ; 
      zoom.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_Z,0)) ;  // InputEvent.META_DOWN_MASK 
      zoom.addActionListener(treePane) ; // action goes there, not here
      zoom.setActionCommand("zoom") ;
      optionsMenu.add(zoom) ; 
      /* original intent was to "zoom" with auto enlarge, no enlarge now*/
      ///// SHIFT L/M/R
      JMenu splitMenu = new JMenu("move split") ; 
      JMenuItem shiftL = new JMenuItem("far left") ;
      shiftL.createToolTip() ; 
      shiftL.setToolTipText("<html><font size=\"+0\">move split far left") ; 
      shiftL.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_LEFT,keyAccelerator)) ;  // InputEvent.META_DOWN_MASK
      shiftL.addActionListener(this) ; 
      shiftL.setActionCommand("left") ;
      splitMenu.add(shiftL) ; 
      JMenuItem shiftM = new JMenuItem("middle") ;
      shiftM.createToolTip() ; 
      shiftM.setToolTipText("<html><font size=\"+0\">center split") ; 
      shiftM.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_DOWN,keyAccelerator)) ;  // InputEvent.META_DOWN_MASK
      shiftM.addActionListener(this) ; 
      shiftM.setActionCommand("middle") ;
      splitMenu.add(shiftM) ; 
      JMenuItem shiftR = new JMenuItem("far right") ;
      shiftR.createToolTip() ; 
      shiftR.setToolTipText("<html><font size=\"+0\">move split far right to search") ; 
      shiftR.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_RIGHT,keyAccelerator)) ;  // InputEvent.META_DOWN_MASK
      shiftR.addActionListener(this) ; 
      shiftR.setActionCommand("right") ;
      splitMenu.add(shiftR) ; 
      optionsMenu.add(splitMenu) ; 

      JMenuItem killer = new JMenuItem("Kill") ;
      killer.createToolTip() ; 
      killer.setToolTipText("<html><font size=\"+0\"kill spawned colog") ; 
      killer.addActionListener(this) ; 
      killer.setActionCommand("killer") ;
      if (this.cologProcessNumber == 0) optionsMenu.add(killer) ; // only 0th

      JMenuItem settings = new JMenuItem("save settings") ;
      settings.addActionListener(this) ; 
      settings.setActionCommand("save_settings") ;
      optionsMenu.add(settings) ; 

      //optionsMenu.addSeparator() ; 
      //JMenuItem save_colog = new JMenuItem("save fol-to-colog") ;  // decided against
      //save_colog.addActionListener(this) ; 
      //save_colog.setActionCommand("save_colog") ;
      //optionsMenu.add(save_colog) ; 
      /////
      optionsMenu.addSeparator() ; 
      // TABLE INSPECTION DIALOG
      JMenuItem inspect_table = new JMenuItem("inspect table") ; 
      inspect_table.setFont(new Font("sanserif",Font.BOLD+Font.ITALIC,12)) ; 
      inspect_table.setForeground(Color.BLACK) ; 
      inspect_table.addActionListener(this) ; 
      inspect_table.setActionCommand("inspect_table") ; 
      optionsMenu.add(inspect_table) ; 

      mb.add(optionsMenu) ; 

      // AUX menu
      JMenu auxMenu = new JMenu("aux") ; 
      auxMenu.setBackground(new Color(255,255,204)) ; 
      // PROOF
      JMenuItem proof = new JMenuItem("extract proof from tree") ; 
      proof.addActionListener(this) ; 
      proof.setActionCommand("proof") ; 
      auxMenu.add(proof) ; 
      // Log COUNTER MODELS 
      JMenuItem cmodels = new JMenuItem("extract counter models from tree") ; 
      cmodels.addActionListener(this) ;
      cmodels.setActionCommand("cmodels") ;
      auxMenu.add(cmodels) ; 
      // Show theory analisis
      JMenuItem analyze = new JMenuItem("theory analysis") ; 
      analyze.addActionListener(this) ; 
      analyze.setActionCommand("analyze") ; 
      auxMenu.add(analyze) ; 
      //  LaTeX PICTURE code
      JMenuItem latex = new JMenuItem("Generate LaTeX picture code for tree") ; 
      latex.addActionListener(this) ;
      latex.setActionCommand("latex") ;
      auxMenu.add(latex) ; 
      mb.add(auxMenu) ; 

/////////////////////////////////////////////////////////////////
///// TOOLBAR ///////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////
      toolbar = new JToolBar() ; 
      toolbar.setBackground(shade) ; 

      Class clazz = this.getClass() ; 
      int fontsize = 24 ; 

      // NAME LABEL
      String toshow = (cologProcessNumber==0 ? "   " : "#"+String.valueOf(cologProcessNumber)) ;
      JLabel nameLabel = new JLabel(" colog"+toshow) ; 
      nameLabel.setFont(new Font(fontName,Font.BOLD+Font.ITALIC,fontsize) ) ; 
      nameLabel.setForeground(Color.BLACK) ; 
      toolbar.add(nameLabel) ;  
      toolbar.add(new JLabel("     ")) ;
      /////
      //RUN button  
      runButton = new JButton("=> ") ; 
      Font runfont = runButton.getFont() ; // Use the system font
      runButton.setFont(new Font(runfont.getFontName(),runfont.getStyle(),16)) ; 
      runButton.setBackground(shade) ; 
      runButton.createToolTip() ; 
      runButton.setToolTipText("<html><font size=\"+0\">   RUN ... build tree  ... ") ; 
      runButton.addActionListener(this) ; 
      runButton.setActionCommand("run") ;
      toolbar.add(runButton) ; 
      // BOUNDS LABEL
      boundsLabel = new JLabel(this.boundsString()) ; 
      toolbar.add(boundsLabel) ; 
      boundsLabel.setMaximumSize(new Dimension(100,30)) ; 
      boundsLabel.addMouseListener(
         new MouseAdapter() {
            public void mouseClicked(MouseEvent e) { showBounds() ;  }
         } ) ;
      boundsLabel.createToolTip() ; 
      boundsLabel.setToolTipText("<html><font size=\"+0\">   Click here to modify bounds ... ") ; 
      /////
      //CLEAR button  
      clearButton = new JButton("clear") ; 
      Font clearfont = clearButton.getFont() ; // Use the system font
      clearButton.setFont(new Font(clearfont.getFontName(),clearfont.getStyle(),16)) ; 
      clearButton.setBackground(shade) ; 
      clearButton.createToolTip() ; 
      clearButton.setToolTipText("<html><font size=\"+0\">      CLEAR tree") ; 
      clearButton.addActionListener(this) ; 
      clearButton.setActionCommand("clear") ;
      toolbar.add(clearButton) ; 
      /////
      toolbar.add(new JLabel("     ")) ;  // separator
      /////
      complexityButton = new JToggleButton("cut") ; 
      Font f = complexityButton.getFont() ; // Use the system font
      complexityButton.setFont(new Font(f.getFontName(),f.getStyle(),16)) ; 
      complexityButton.setBackground(shade) ; 
      complexityButton.setForeground(Color.GRAY) ; 
      complexityButton.createToolTip() ; 
      complexityButton.setToolTipText("<html><font size=\"+0\">      CUT function complexity is OFF") ; 
      complexityButton.addActionListener(this) ; 
      complexityButton.setActionCommand("complexity") ; 
      /////
      // CSPINNER
      cspinner = new JSpinner(new SpinnerNumberModel(0,0,50,1)) ; 
      cspinner.createToolTip() ; 
      cspinner.setMaximumSize(new Dimension(50,30)) ; 
      cspinner.setToolTipText("<html><font size=\"+0\">        max operator nesting depth  ") ; 
      cspinner.addChangeListener(this) ; 
      /////
      toolbar.add(complexityButton) ; 
      //cspinner.setRequestFocusEnabled(false) ;
      toolbar.add(cspinner) ; 
      /////
      toolbar.add(new JLabel("     ")) ;  // separation
      //////////////////////////////////////////////////
      // RULE LABEL at bottom of window
      ruleLabel = new JLabel("      ") ; 
      ruleLabel.setForeground(Color.BLUE) ; 
      ruleLabel.setFont(new Font(fontName,Font.PLAIN,16) ) ; 
      ruleLabel.createToolTip() ; 
      ruleLabel.setToolTipText("   proposed bindings") ; 
      treePane.info = ruleLabel ; 
      treePane.auxArea = this.auxArea ;  // 12.13.2015

      // THEORY PANE
      theoryPane = new JList<String>(theoryListModel) ; 
      theoryPane.setSelectionForeground(Color.BLUE) ;  // inference ~ blue
      theoryPane.setSelectionBackground(Color.lightGray) ;  // needs be light
      theoryPane.createToolTip() ; 
      String os = System.getProperty("os.name") ; 
      // modified 9/1/2009
      //if ( true || os.indexOf("Mac") > -1 || os.indexOf("Linux" ) > -1) { 
         theoryPane.createToolTip() ; 
         String about_drop = 
            "<html><font size=\"+0\"><b>COLOG THEORY PANE</b><br>&nbsp;DROP theory here  to open it (" + os + ").<br>" +
            "&nbsp;[n] means n<i>th</i> rule is a queued rule. <br>&nbsp;Existential variables are <font color = \"RED\">RED</font>.<br>"+
            "&nbsp;Click rule here to view all inference nodes for the rule in the tree.&nbsp;&nbsp; <br> "    ;   
         theoryPane.setToolTipText(about_drop) ; 
      //}
      theoryPane.setTransferHandler(new DropHandler()) ; 
      theoryPane.setFont(new Font((os.indexOf("Mac") > -1 ? "monospaced":"monospaced"),Font.PLAIN,14) ) ; 
      theoryPane.setBorder(BorderFactory.createTitledBorder("colog theory"));
      ////
      theoryPane.getSelectionModel().addListSelectionListener(this) ;
      JScrollPane theoryScroller = new JScrollPane(theoryPane) ; 
      // Tell treePane about theoryPane for rule selections
      treePane.rulelist = theoryPane ;
      ///// 
      splitPane = 
         //new JSplitPane(JSplitPane.HORIZONTAL_SPLIT,tabbedPane,theoryScroller) ; //xxxxx
         new JSplitPane(JSplitPane.HORIZONTAL_SPLIT,theoryScroller,tabbedPane) ; //reversed ???
      this.getContentPane().add("North", toolbar) ; 
      this.getContentPane().add("Center", splitPane) ; 
      this.getContentPane().add("South", ruleLabel) ; 
      int halfwit = this.getSize().width/2 ;
      splitPane.setDividerLocation(halfwit) ; // double seems not to work

      this.addWindowListener(new WindowAdapter() { 
         public void windowClosing(WindowEvent e) { 
            exitExam() ; // confirm only ???
         }
         // iconify tools when this is iconified
         //public void windowIconified(WindowEvent w) { 
         //   treePane.selected = -1 ; // deselect
         //}
         // deiconify tools when this is deiconified
         //public void windowDeiconified(WindowEvent w) {
         //}
      }) ; 

      Thread shows = new Thread() { 
         public void run() { 
            setVisible(true) ;
         }
      } ;
      EventQueue.invokeLater(shows); 
      this.toggle = true ;  
      /////
      redirectSystemStreams() ;
   } // end Colog() constructor


///////////////////////////////////////////////////////////////////

   private String boundsString() { 
      return 
        "<html>depth="+String.valueOf(tree.depth) +
        "<br>width="+String.valueOf(tree.width)  ;
   }

   private void updateConsole(final String text) {  
      SwingUtilities.invokeLater(new Runnable() {  
        public void run() {  
          consoleArea.append(text);  
        }  
      });  
   }  

   private void redirectSystemStreams() {  
      OutputStream out = new OutputStream() {  
         @Override  
         public void write(int b) throws IOException { 
            updateConsole(String.valueOf((char) b));  
         }  
   
         @Override  
         public void write(byte[] b, int off, int len) 
                                      throws IOException {  
            updateConsole(new String(b, off, len));  
         }  

         @Override  
         public void write(byte[] b) throws IOException {  
            write(b, 0, b.length);  
         }  
      };  
      System.setOut(new PrintStream(out, true));  
      System.setErr(new PrintStream(out, true));  
   }  
///////////////////////////////////////////////////////////////////

   /**
     *  Update this using the current tree. 
     */
   protected void update() { 
      this.orient() ;
   }

   /**
     *  Orient the tree view so that scrollbars are visible if needed.
     *  The getPreferredSize() method of TreePane needs be 
     *  a little bigger than expected(?).
     */
   protected void orient() { 
      treePane.setXY() ; // NEEDED because repaint will delay too long
      treePane.repaint() ; 
      this.wiggle() ; 
   }

   //
   // wiggle the divider, in case scrollbars are needed ;-) kludge 
   //
   protected void wiggle() { 
      this.splitPane.setDividerLocation(splitPane.getDividerLocation() + (toggle?1:-1)) ;
      if (toggle) toggle = false; else toggle = true ; 
      // keep focus in view
      JScrollBar v = this.treeScroller.getVerticalScrollBar() ,
                 h = this.treeScroller.getHorizontalScrollBar() ;  
      Dimension d = treePane.getPreferredSize() ; 
      v.setValue(d.width) ; 
      h.setValue(d.height) ; 
   }

   /**
     *  Adjust bounds, and possibly tree. 
      */
   protected void setBounds(int d, int w) { 
      this.clear() ; 
      if (d != tree.depth) { 
         // save do rule if necessary
         Rule doRule = null ; 
         if (tree.activeTheory != null)
            doRule = tree.activeTheory.removed ; // may still be null
         tree = new Tree(d,w) ;  // new tree
         tree.setLogging(true) ;   // logging true when using GUI
         if (this.theory != null) { 
            if (doRule != null) this.theory.add(0,doRule) ; 
            tree.setTheory(this.theory) ; 
            this.setTheoryList(this.theory) ; // will warn!
         } 
         treePane.setCologTree(tree) ; 
         treePane.setXY() ;
         
      }
      else { 
         tree.width = w ;  //same tree, new width 
         tree.capacity = d * w ; 
         tree.setLogging(true) ;
      }
      treePane.setCologTree(tree) ; 
      if (complexityButton.isSelected())
         tree.complexity_limit = ((Integer)cspinner.getValue()).intValue() ;
      boundsLabel.setText(this.boundsString()) ; 
   }
   
   // encapsulate bounds inquiry 
   protected void showBounds() {
        Object[] contents = new Object[4] ; 
        contents[0] =  new JLabel("depth") ;  
        contents[1] =  new JTextField() ; 
        contents[2] =  new JLabel("width") ;  
        contents[3] =  new JTextField() ; 
        Object[] options = { "Ok", "Cancel" } ; 
        boolean error = true ;  // loop until same or sensible
        while (error) { 
           ((JTextField)contents[1]).setText(String.valueOf(this.treeDepth)) ; 
           ((JTextField)contents[3]).setText(String.valueOf(this.treeWidth)) ; 
           int result =    
              JOptionPane.showOptionDialog(this ,
                                          contents, 
                                          "    Set Search Bounds    ", 
                                          JOptionPane.DEFAULT_OPTION, 
                                          JOptionPane.QUESTION_MESSAGE,
                                          null, options, options[0]); 

           int d = -1, w = -1 ; 
           try { 
              d = Integer.parseInt(((JTextField)contents[1]).getText()) ; 
              w = Integer.parseInt(((JTextField)contents[3]).getText()) ; 
              error = false ; 
           }
           catch(Exception xx) {} // error ...
           if (d >= 1 && w >= 1) { 
              this.treeDepth = d ; 
              this.treeWidth = w ;
              error = false ; 
           } else error = true ; 
        } // end while(error) 
        this.setBounds(this.treeDepth,this.treeWidth) ; 
   }

   // encapsulate bounds inquiry 
   protected void showBounds2() {  // 4 October 2014  ABANDONED
        JPanel pane =  new JPanel() ; 
        pane.setLayout(new GridLayout(3,3)) ; 
        /*          size    fixed       0   1   2
            depth                       3   4   5
            width                       6   7   8
        */
        Component[] comps = new Component[9] ; 
        comps[0]    =  new JLabel("   ") ;  
        comps[1]    =  new JLabel(" size") ;  
        comps[2]    =  new JLabel("fixed") ;  
        comps[3]    =  new JLabel("depth") ;  
        comps[4]    =  new JTextField() ;   // DEPTH
        comps[5]    =  new JRadioButton() ;  
        comps[6]    =  new JLabel("width") ;  
        comps[7]    =  new JTextField() ;   // WIDTH
        comps[8]    =  new JRadioButton() ;  
        for (int i = 0 ; i < 9 ; i++) pane.add(comps[i]) ;
        Object[] options = { "Ok", "Cancel" } ; 
        boolean error = true ;  // loop until same or sensible
        while (error) { 
           ((JTextField)comps[4]).setText(String.valueOf(this.treeDepth)) ; 
           ((JRadioButton)comps[5]).setSelected(tree.depth_fixed) ; 
           ((JTextField)comps[7]).setText(String.valueOf(this.treeWidth)) ; 
           ((JRadioButton)comps[8]).setSelected(tree.width_fixed) ; 
           int result =    
              JOptionPane.showOptionDialog(this ,
                                          pane, 
                                          "    Set Search Bounds    ", 
                                          JOptionPane.DEFAULT_OPTION, 
                                          JOptionPane.QUESTION_MESSAGE,
                                          null, options, options[0]); 

           int d = -1, w = -1 ; // BAD
           boolean dfixed = false, wfixed = false ;
           try {  
              d = Integer.parseInt(((JTextField)comps[4]).getText()) ; 
              w = Integer.parseInt(((JTextField)comps[7]).getText()) ; 
              dfixed = ((JRadioButton)comps[5]).isSelected() ; 
              wfixed = ((JRadioButton)comps[8]).isSelected() ; 
              error = false ;  // still bad
           }
           catch(Exception xx) {} // error ...
           if (d >= 1 && w >= 1) { 
              this.treeDepth = d ; 
              this.treeWidth = w ; 
              //tree.depth_fixed = dfixed ;   // later
              //tree.width_fixed = wfixed ;
              tree.depth_fixed = true ;
              tree.width_fixed = true ;
              error = false ; 
           } else error = true ; 
        } // end while(error) 
        this.setBounds(this.treeDepth,this.treeWidth) ; 
   }



   // encapsulate table dialog inquiry 
   protected void tableDialog() {
        Object[] contents = new Object[3] ; 
        contents[0] =  new JLabel("<html>Inspect table for") ;  
        contents[1] =  new JTextField() ; 
        ((JTextField)contents[1]).setBorder(BorderFactory.createTitledBorder("branch #")) ;
        contents[2] =  new JLabel("<html>Console Report appears<br> <font color=\"RED\"><i>after</i></font> next run.") ;  
        //contents[3] = new JRadioButton("fact table") ;  
        //contents[4] = new JRadioButton("subfact table") ;
        //ButtonGroup grp = new ButtonGroup() ; 
        //grp.add((JRadioButton)contents[3]) ; 
        //grp.add((JRadioButton)contents[4]) ; 
        //((JRadioButton)contents[3]).setSelected(true) ;
        Object[] options = { "Ok", "Cancel" } ; 
        int b = -1 ; 
        int result =    
              JOptionPane.showOptionDialog(this ,
                                          contents, 
                                          "   Table Inspector    ", 
                                          JOptionPane.DEFAULT_OPTION, 
                                          JOptionPane.QUESTION_MESSAGE,
                                          null, options, options[0]); 

        try { 
           if (result == 0)
              b = Integer.parseInt(((JTextField)contents[1]).getText()) ; 
        }
        catch(Exception xx) {  ; } 
        //System.out.println(b) ; 
        if (b > 0) {
           tree.atLeaf = b ; 
           System.out.println("Inspect table for branch #"+b+" f=factual s=subfactual") ; 
        }
        else  { 
           tree.atLeaf = -1 ;
           System.out.println("No table inspection." ) ; 
        }
   }

   /**
     *  Actions.
     */
   public void actionPerformed(ActionEvent e) { 
      String command = e.getActionCommand() ; 
      // System.out.println(command) ;       
      if (command.equals("Toolbar")) {
         if (toolbar.isVisible()) { 
            toolbar.setVisible(false) ; 
         }
         else { 
            toolbar.setVisible(true) ; 
         }
      }   
      ////
      else if (command.equals("open")) open() ; 
      else if (command.equals("reload")) {
         if (!theoryDescription.equals(""))  // no current
            open(theoryDescription) ; 
      }
      else if (command.equals("openURL")) {   
         Object[] contents = new Object[3] ; 
         contents[0] =  new JLabel("enter, drop or paste URL:                            ") ;  
         contents[1] =  new JTextField() ; 
         contents[2] =  new JLabel("colog will attempt to load this theory ...") ; 
         Object[] options = { "Ok", "Cancel" } ; 
         int result =    
            JOptionPane.showOptionDialog(this ,
                                       contents, 
                                       "    Open URL    ", 
                                       JOptionPane.DEFAULT_OPTION, 
                                       JOptionPane.QUESTION_MESSAGE,
                                       null, options, options[0]); 
         if (result == 0) { 
            String input = ((JTextField)contents[1]).getText() ;
            if (!input.equals("")) open(((JTextField)contents[1]).getText()) ;
         }
      } 
      else if (command.equals("latex")) latex() ;
      else if (command.equals("cmodels")) cmodels() ;
      else if (command.equals("analyze")) analyze() ;
      else if (command.equals("run")) { 
         ruleLabel.setText("      ") ; // does not clear text first
         this.autorun() ; 
      }
      else if (command.equals("clear")) { 
         this.clear() ; 
         ruleLabel.setText("      ") ; 
      }
      else if (command.equals("bounds")) {  
        this.showBounds2() ;   
      } 
      else if (command.equals("proof")) { 
            //DataWindow proofWindow = new DataWindow() ; 
            //proofWindow.setLocation(400,400) ; 
            //proofWindow.setSize(600,400) ; 
            //proofWindow.setTitle("Proof") ; 
            auxArea.setText(this.theoryDescription+"\n"+this.tree.proof()) ; 
            //proofWindow.setVisible(true) ; 
      }
      else if (command.equals("new")) {  
         File src = new File(System.getProperty("user.dir")+File.separator+CologConstants.JARNAME) ; // N.B. wrong name for test
         System.out.println(src) ;
         if (!src.exists()) 
            JOptionPane.showMessageDialog(this,"<html>Put a copy of <b>"+CologConstants.JARNAME+"</b><p> in the directory <tt>"+System.getProperty("user.home"),"Cannot find "+CologConstants.JARNAME,JOptionPane.INFORMATION_MESSAGE) ;
         else {
            this.newCologProcess() ; 
         }
      }
      else if (command.equals("complexity")) {  
         if (complexityButton.isSelected()) {
            tree.complexity_limit = ((Integer)cspinner.getValue()).intValue() ;
            complexityButton.setToolTipText("<html><font size=\"+1\">        CUT function complexity is ON") ;
            complexityButton.setForeground(Color.RED) ; 
            complexityButton.setBackground(Color.WHITE) ; 
         }
         else {
            tree.complexity_limit = -1 ; // OFF
            complexityButton.setToolTipText("     CUT function complexity is OFF") ;
            complexityButton.setForeground(Color.GRAY) ; 
         }
      }
      else if (command.equals("left")) {  
         splitPane.setDividerLocation(0.0) ;
      }
      else if (command.equals("middle")) {  
         splitPane.setDividerLocation(0.5) ;
      }
      else if (command.equals("right")) {  
         splitPane.setDividerLocation(1.0) ;
      }
      else if (command.equals("killer")) 
         kill() ;
      else if (command.equals("save_settings")) { 
         Point p = this.getLocation() ; 
         settings.put("x",String.valueOf(p.x)) ; 
         settings.put("y",String.valueOf(p.y)) ; 
         Dimension d = this.getSize() ; 
         settings.put("w",String.valueOf(d.width)) ; 
         settings.put("h",String.valueOf(d.height)) ; 
         //settings.put("command",this.command) ; 
         try { 
            FileOutputStream fos = 
               new FileOutputStream(
                 System.getProperty("user.home")+File.separator+"colog.settings") ; 
            settings.store(fos, "colog user properties") ; 
         }
         catch(Exception x) {  
            System.out.println(x) ; 
         }
         JOptionPane.showMessageDialog(this,
             "<html><b>Saved <blockquote> size<br>location on screen<br></blockquote>") ; 
      }
      else if (command.equals("inspect_table")) { tableDialog() ;  }
      else this.orient() ; 
      /////
   }  // end actionPerformed

   public void stateChanged(javax.swing.event.ChangeEvent e) { 
      if (complexityButton.isSelected())
         tree.complexity_limit = ((Integer)cspinner.getValue()).intValue() ;
   }

   /**
     * encapsulate new colog process with dialog
      */
   protected void newCologProcess() {
        //String command = "java -Xms6g -Xmx10g -jar " + 
        //   System.getProperty("user.home") + 
        //   System.getProperty("file.separator") +   
        //  CologConstants.JARNAME ; 
        Object[] contents = new Object[4] ; 
        contents[1] =  new JLabel("Edit command if necessary...") ;  
        ((JLabel)contents[1]).setForeground(Color.RED) ; 
        contents[2] =  new JLabel("See console on error.") ;  
        contents[0] =  new JTextField(this.command) ; 
        Object[] options = { "Ok", "Cancel" } ; 
        int result =    
           JOptionPane.showOptionDialog(this ,
                               contents, 
                               "    New colog GUI process    ", 
                               JOptionPane.DEFAULT_OPTION, 
                               JOptionPane.QUESTION_MESSAGE,
                               null, options, options[0]); 
        if (result == 1) return ;  //  otherwise 0:
        try { 
            //this.command =  ((JTextField)contents[0]).getText() ;
            //command+=" "+ String.valueOf(processes.size()+1) ;  
            String todo = this.command +" "+ String.valueOf(processes.size()+1) ;  
            String[] cmdarray = {todo} ;  // 2023/1/28
            final Process p = 
                    Runtime.getRuntime().exec(cmdarray) ; 
            processes.add(p) ; alive.add(true) ; 
        }
        catch(Exception xx) {  System.out.println(xx) ; } // error ...
   }


   // Present list of processes for user to kill
   protected void kill() {
      Object[] choose = new Object[processes.size()] ;
      for (int p = 0 ; p < processes.size() ; p++)  {
         choose[p] = new JRadioButton("colog #"+String.valueOf(p+1)) ;  
         if (!alive.elementAt(p)) ((JRadioButton)choose[p]).setEnabled(false) ;
      }
      Object[] options = { "Kill", "Cancel" } ; 
      int result =    
         JOptionPane.showOptionDialog(this ,
                                       choose, 
                                       "Kill process # ...", 
                                       JOptionPane.DEFAULT_OPTION, 
                                       JOptionPane.QUESTION_MESSAGE,
                                       null, options, options[0]); 
      try { 
         for (int c = 0 ; c < choose.length ; c++) {
            if (((JRadioButton)choose[c]).isSelected()) {
                Process p = processes.elementAt(c) ; 
                //process.remove(process.elementAt(k)) ;  
                p.destroy() ;
                alive.set(c,false) ; 
             }
          }  
        }
        catch (Exception xx) { System.out.println(xx) ; }
   }
   
   // GUI stub for testing ideas
   private void test() { 
      // send report to aux
   }

   /**
     *  autorun from GUI -- This could be a Threaded with a run
     */
   protected void autorun() { 
      this.clear() ; 
      ruleLabel.setText("      ") ; 
      if (this.theory == null) return ; 
      long time =  System.currentTimeMillis() ; // milliseconds 
      tree.dfSearch() ;   // tree.run()  ?
      time = System.currentTimeMillis() - time ; 
      // report on the run 
      String message = "no message yet" ; 
      // Report to console
      System.out.println("depth="+tree.depth+" width="+tree.width) ; 
      System.out.println("goal Branches     = " + tree.goalBranches) ; 
      System.out.println("false Branches    = " + tree.falseBranches) ; 
      System.out.println("Models            = " + tree.counterModels) ; 
      System.out.println("Too Deep Branches = " + tree.tooDeepBranches) ; 
      System.out.println("Too Wide Branches = " + tree.tooWideBranches) ; 
      if (tree.counterModels.isEmpty() && tree.tooDeepBranches.isEmpty() && tree.tooWideBranches.isEmpty()) 
         message = "PROOF" ; 
      else if (!tree.counterModels.isEmpty())  {
         String s = (tree.counterModels.size() == 1 ? "" : "s") ; 
         if (complexityButton.isSelected()) 
            message = "Possible model"+s+", see Colog Console" ; 
         else
            message = "Model"+s+", see Colog Console" ; 
      }
      else 
         message = "Blocked branches, see Colog Console" ;
      this.update() ;
      ruleLabel.setForeground((Color.MAGENTA).darker().darker()) ;  
      String report = "#inferences=" + String.valueOf(tree.inferenceTally) + ", #facts=" + String.valueOf(tree.count) +", #branches=" + String.valueOf(tree.numberOfBranches())+", deepest=" + String.valueOf(tree.descent)  +", time="+String.valueOf(time)+" ms " ;
      System.out.println(report+"#subsumed="+tree.subsumed+ " #irrelevant_branches=" + tree.unused_branch_count) ; 
      //System.out.println(this.theory) ; 
      System.out.println("-------------------------------------------") ; 
      ruleLabel.setText(message+", "+ report); 
      this.wiggle() ;  // one more wiggle ;-) attention scrollers 
      // HARD RESET -- RECONSTRUCT 
      if (this.theory !=null) {  // hard reset
         if (tree.activeTheory.removed != null) 
            this.theory.add(0,tree.activeTheory.removed) ; 
         tree.setTheory(this.theory) ; 
      }
   }


   /**
     *  Clear the tree.
     */
   protected void clear() { 
      // initialize tally array 
      treePane.setSelected(-1) ; // de-select actually 
      //treePane.dw.setVisible(false) ; 
      tree.clear() ; 
      this.auxArea.setText("") ; // 
      this.consoleArea.setText("") ; // 29 May 2014
      this.update() ;
   }

   /**
     *  Open colog theory via file dialog.
     */
   protected synchronized void open() {
      JFileChooser fileChooser = new JFileChooser() ;  
      /////
      // attempt to start dialog in previous/last directory..
      String fs = null ; 
      if (theoryDescription.startsWith("file:"+File.separator+File.separator+"localhost")) 
         fs = theoryDescription.substring(16) ;
      else if (theoryDescription.startsWith("file:")) 
         fs = theoryDescription.substring(5) ; 
      if (fs != null) fileChooser.setSelectedFile(new File(fs)) ;   
      /////
      int result = fileChooser.showOpenDialog(this) ; 
      if (result == JFileChooser.APPROVE_OPTION) {    
         this.open(fileChooser.getSelectedFile().toURI().toString()) ; 
      }      
   } // end open()


   /**
     *  Generate LaTex for picture tree
     */
   protected synchronized void latex() {
      //File f = new File(theorySpec.getAbsolutePath() + ".tex") ; 
      LaTeXTree latextree = new LaTeXTree(this.treePane) ; 
      // inquire regarding x offset for tree
      String s = 
         JOptionPane.showInputDialog(this,"Tree x-offset (Cancel uses offset=0): ", "0") ;
      int xoff = 0 ;            
      try{ 
         xoff = Integer.parseInt(s) ; 
      }
      catch (Exception xxx) { System.out.println(xxx) ; }
      latextree.xoffset = xoff ;  // 0 if there was some error
      if (!theoryDescription.equals(""))
         auxArea.setText(latextree.laTeX(theoryDescription)) ;
   } // end latex()

   /**
     *  Write counter model(s) to auxArea
     */
   protected synchronized void cmodels() { 
      if (tree.counterModels == null || tree.counterModels.isEmpty()) 
         auxArea.setText("No counter models") ; 
      else auxArea.setText(tree.reportCounterModels()) ;
   } // end cmodels()

   /**
     *  Write theory analysis to auxArea
     */
   protected synchronized void analyze() { 
      if (this.theory == null) 
         auxArea.setText("No theory loaded.") ; 
      else auxArea.setText(this.theoryDescription+"\n"+CologAnalyzer.report(this.theory)) ;
   } // end analyze()

   /**
     *  Add theory to DefaultListModel (and 'no match' item).
     */
   protected void setTheoryList(Theory theory) {
      this.theoryListModel.clear() ; 
      Iterator it = theory.iterator() ; 
      int c = 1 ; 
      while (it.hasNext()) { 
         Rule rule = (Rule)it.next() ; 
         // Mark existential rules as "declarative" ! ... 
         String end = ((rule.existential>0) ? "?" : ".") ; 
         String l = 
            (tree.activeTheory.qued(c) ? "[" : " ") ;  
         String r = 
            (tree.activeTheory.qued(c) ? "] " : "  ") ;  
         //String s =l+c+r + rule.toString() + end ;
         //String s ="<html><font color=\"RED\" size=\"+2\">"+ l+c+r + rule.toString() + end ;
         this.theoryListModel.addElement(rule.toHTML(tree.activeTheory.qued(c),c,4,rule.U)) ; 
         //System.out.println("#  "+s) ; 
         c++ ; 
      }
      // add bottom-out choice
      this.theoryListModel.addElement(NOMATCH) ; 
   }

   /**
     *  Before exit ...
     *  Changes are not cl osely tracked.  
     */
   private void exitExam() {                     
      Toolkit.getDefaultToolkit().beep() ; 
      Object[] options = {"Quit","Cancel"} ;
      int choice = 
         JOptionPane.showOptionDialog(this, 
                               " quit?",
                               "Confirm Quit", 
                               JOptionPane.YES_NO_OPTION, 
                               JOptionPane.WARNING_MESSAGE, 
                               null, 
                               options, options[0]  ); 
      if (choice == 0)  System.exit(0) ;  
   }   

   /**
     *  GUI from the command line.
     */
   public static void main(String[] args) { 
         //System.out.println(args.length) ;
         Colog gui = new Colog(0) ; 
         if (args.length == 1) gui.open(args[0]) ;
         // e.g., args[0].equals  file:/colog14G/co/dpe.gl
   }

   /**
     * Extract the String flavor, which gives a reasonable
     * URL description for Java.  Need cross-platform flavor. 
     * Perhaps add DataFlavor.stringFlavor attempt to importData?
     * inner
      */
   class DropHandler extends TransferHandler { 

       public boolean canImport(JComponent comp, DataFlavor[] flavor) {
          return true ;  // 
       }

       public boolean importData(JComponent comp, Transferable t) { 
         DataFlavor[] flavors = t.getTransferDataFlavors() ; 
         int p = -1 ; 
         String s = "java.awt.datatransfer.DataFlavor[mimetype=text/uri-list;representationclass=java.lang.String]" ;
         for(int i = 0 ; i < flavors.length ; i++) {
            //System.out.println(flavors[i].toString()) ; // inspect the flavors
            if (flavors[i].toString().startsWith(s)) p = i ;
         }
         if (p != -1)  {   // use this String representation  Mac or Linux
            try { 
               String u = ((String)t.getTransferData(flavors[p])).trim() ; 
               System.out.println(u) ; 
               open(url(u)) ; 
               // open(new File(u).toURL().toString()) ;  // ??? alternate
               return true ;  
            } catch(Exception x1) { System.out.println(x1) ; }
         }  // end if
         else if (t.isDataFlavorSupported(DataFlavor.javaFileListFlavor))  {
            try { 
               java.util.List list = (java.util.List)t.getTransferData(DataFlavor.javaFileListFlavor);
               Iterator it = list.iterator() ; 
               //System.out.println(DataFlavor.javaFileListFlavor.toString()) ; 
               File file = (File)it.next() ;  // it is a FILE
               open(file.toURI().toURL().toString()) ;  
               return true ; 
            } catch (Exception x2) { System.out.println(x2) ; }
         }  // end else if 
         // otherwise ???
         return false ;
      }

   } // end inner DropHandler

   // a work-around for Java 1.8  ?
   private static String url(String s) { 
      try { 
         return new URL(s).toString() ; 
      }
      catch (Exception x) {
         return "file:"+s ; 
      }
   }

   /**
     * open theory linked to url
     */
   public void open(String U) {  // can be called via command-line
      this.folTool.clear() ;
      URL url = null ;
      try { 
         url = new URL(U) ; 
      }
      catch (Exception mue) { 
         System.out.println(mue) ;  // see console
         JOptionPane.showMessageDialog(this,
             "<html><b>Bad drop<br>See colog Console for report.") ; 
      }
      if (url == null) return ; // leave GUI as was 
      // get url
      Theory xtheory = null ; // null return if exception 
      boolean trouble = false ;
      try { 
         InputStream in = url.openStream() ; 
         cologLoader loader = new cologLoader(in) ;    
         if (loader.theory == null || loader.theory.size()==0 || loader.error) 
            trouble = true ; 
         xtheory = loader.theory ;  // could be null
         in.close() ;
      }
      catch (Exception x) { 
         trouble = true ; 
      }
      if (trouble) { 
         JOptionPane.showMessageDialog(this,
             "<html><b>Load errors<br>See console.") ; 
         return ; // leave current theory alone
      }
      // otherwise install new xtheory
      // clear all displays
      this.clear() ; 
      this.consoleArea.setText("") ; 
      //this.tree.clear() ; 
      //this.setTitle("Colog") ; 
      this.theoryListModel.clear() ; 
      this.ruleLabel.setText("   ") ; 
      this.treePane.selected =  -1 ; // de-select
      this.theory = xtheory ;
      this.theoryDescription = url.toString() ;
      this.setTitle(BANNER + this.theoryDescription) ; 
      tree.setTheory(this.theory) ; 
      this.setTheoryList(this.theory) ;
      this.update() ; 
      this.useHints() ;
   }

   /**
     * Attempt to use any hints provided in the theory,
     * or revert to default depth (D) and width (W).
      */
   protected void useHints() { 
      this.treeWidth = W ;
      this.treeDepth = D ; 
      if (this.theory.isFlat()) this.treeWidth = 1 ; 
      Iterator it = this.theory.hints.keySet().iterator() ; 
      boolean cutflag = false ;  // unless hint turns it on
      while (it.hasNext()) { 
         String attr = (String)it.next() ; 
         if (attr.equals("cut")) { 
            try { 
               int c = Integer.decode(this.theory.hints.get(attr)) ; 
               consoleArea.setText(consoleArea.getText()+"# hint: cut="+c+"\n") ; 
               if (c > 0) { 
                  complexityButton.setSelected(true) ; 
                  complexityButton.setForeground(Color.RED) ; 
                  tree.complexity_limit = c ; 
                  cspinner.setValue(c) ;
                  complexityButton.setToolTipText("<html><font size=\"+1\">        CUT function complexity is ON") ;
                  cutflag = true ; 
               }
               else { 
                  complexityButton.setSelected(false) ; // 6.23.2014
                  tree.complexity_limit = -1 ; // OFF
                  complexityButton.setToolTipText("     CUT function complexity is OFF") ;
                  complexityButton.setForeground(Color.BLACK) ; 
                  cspinner.setValue(0) ;
               }
            } catch (Exception cx) {}  // cannot parse int
         } // end parse cut
         else if (attr.equals("width")) { 
            try { 
               int w = Integer.decode(this.theory.hints.get(attr)) ; 
               consoleArea.setText(consoleArea.getText()+"# hint: width="+w+"\n") ; 
               if (w > 0) this.treeWidth = w ;
            }
            catch (Exception wx) {}  // not a width 
         } // end parse width
         else if (attr.equals("depth")) { 
            try { 
               int d = Integer.decode(this.theory.hints.get(attr)) ; 
               consoleArea.setText(consoleArea.getText()+"# hint: depth="+d+"\n") ; 
               if (d > 0) this.treeDepth = d ;
            }
            catch (Exception wx) {}  // not a depth 
         } // end parse depth
         //System.out.println(attr+"="+this.theory.hints.get(attr)) ;
      } // end while
      if (!cutflag) { 
                  complexityButton.setSelected(false) ;
                  tree.complexity_limit = -1 ; // OFF
                  complexityButton.setToolTipText("     CUT function complexity is OFF") ;
                  complexityButton.setForeground(Color.BLACK) ; 
                  cspinner.setValue(0) ;
      }
      this.setBounds(this.treeDepth,this.treeWidth) ;
   }

   /**
     *  Rule use display in tree.  Inner class.
     *  Still not as originally intended, but nice this way ;-)
     *  1st click on tree node highlights all nodes using inference rule, 
     *  2nd click limits to just the particular tree node - nice.
      */
   public void valueChanged(ListSelectionEvent e) {
      treePane.rule = theoryPane.getSelectedIndex()+1 ;
      treePane.selected = -1 ; 
      treePane.repaint() ; 
   }

   // message from FOL tool
   protected void setTheory(Theory xtheory) {   
      this.theoryDescription = "" ; // prevents reload 
      this.clear() ; 
      this.consoleArea.setText("") ; 
      this.theoryListModel.clear() ; 
      this.ruleLabel.setText("   ") ; 
      this.treePane.selected =  -1 ; // de-select
      this.theory = xtheory ;
      this.setTitle(BANNER + this.theoryDescription) ; 
      tree.setTheory(this.theory) ; 
      this.setTheoryList(this.theory) ;
      this.update() ; 
      this.useHints() ;
   }

} // end Colog


// about 500 fewer lines of code than at the beginning of the year 2013

   
 
     
     
 
