package sam.gui ;

import java.util.* ; 
import java.awt.* ; 

/**
  *  Generate LaTeX code for a colog tree.
  *  Last modified 16 October2011, {\tt  ...}
  *  Last modified 20 February2013, {\texttt  ...}  
  *  Last modified 1 July 2014, \mathtt{ ...} 
  *  Needs mod for single child (|).
  *  Needs stitching.
  */   
public class LaTeXTree {
   private TreePane treepane ;

   public int xoffset ; 

   /**
     * the allowable LaTeX line slopes (dx,dy) 
     * ordered using < on dy/dx
      */
   private int[] // 26  I think this is all of them.... 
     dx = {1,1,1,1,1,2,1,3,2,3,4,5,1,6,5,4,3,5,2,5,3,0,6,5,4,3} ,
     dy = {0,6,5,4,3,5,2,5,3,4,5,6,1,5,4,3,2,3,1,2,1,1,1,1,1,1} ;
   private double[] slope ; 

   public LaTeXTree(TreePane tp) {
      this.treepane = tp ; 
      this.xoffset = 70 ; //
      slope = new double[26] ;
      for(int i = 0 ; i < 26 ; i++) slope[i] = (double)dy[i]/(double)dx[i] ;
      //for(int i = 0 ; i < 21 ; i++) System.out.println(slope[i]) ;
   }

   /**
     *  index of LaTeX slope for line having slope r.
      */
   private int slope(double r) { 
      int k = 1 ;
      double diff = Math.abs(r-slope[1])  ;  
      for (int i = 2 ; i < 26 ; i++) { 
         double d = Math.abs(r - slope[i]) ;
         if (d < diff) { 
            k = i ; 
            diff = d ; 
         } 
      }
      return k ;
   }
          
   /**
     *  LaTeX for arcs in tree 
      */
   private static String putLine(double x,double y,int dx,int dy,double len) { 
      StringBuffer sb = new StringBuffer() ; 
      sb.append("\\put(") ; 
      sb.append(String.valueOf(x)) ; 
      sb.append(",") ; 
      sb.append(String.valueOf(y)) ; 
      sb.append("){\\line(") ; 
      sb.append(String.valueOf(dx)) ; 
      sb.append(",") ; 
      sb.append(String.valueOf(dy)) ; 
      sb.append("){") ; 
      sb.append(String.valueOf(len)) ; 
      sb.append("}} \n") ; 
      return sb.toString() ; 
   }

   /**
     *  LaTeX code for facts in tree
     */
   private static String putFact(double x,double y,String fact) { 
      StringBuffer sb = new StringBuffer() ; 
      sb.append("\\put(") ; 
      sb.append(String.valueOf(x)) ; 
      sb.append(",") ; 
      sb.append(String.valueOf(y)) ; 
      sb.append("){$\\mathtt{ ") ;    // replace {\texttt ... 
      sb.append(fix(fact)) ;
      sb.append("}$} \n") ;       // ... }$
      return sb.toString() ; 
   }

  /**
    *  Replace "-" with "\\_" except last.
    */
   private static String fix(String fact) {
      int last = fact.lastIndexOf("_") ; 
      if (fact.substring(0,last).contains("_")) {
         StringBuffer sb = new StringBuffer() ; 
         for (int i=0 ; i < last ; i++) {
            String s = fact.substring(i,(i+1)) ;
            if (s.equals("_"))
               sb.append("\\_") ;
            else sb.append(s) ; 
         }
         return sb.toString() ;
      }
      else return fact ; 
  }

   /**
     *  generate LaTeX figure & picture blocks for the tree, 
     *  encased in mini document.
     */
   public String laTeX(String theoryfile) {
      int selected = (treepane.selected > -1 ? treepane.selected : 0) ; 
      if (treepane.tree == null) return "" ;     // no tree
      StringBuffer sb = new StringBuffer() ; 
      // // // // // 
      sb.append("\\documentclass[11pt]{article} \n") ;  // 
      sb.append("\\begin{document} \n\n") ;  
      // // // // //
      sb.append("\\begin{figure}[h] \n") ;
      sb.append("\\setlength{\\unitlength}{0.025cm} \n") ; 
      double H = height(selected)   ;
      sb.append("\\begin{picture}(0,"+ String.valueOf(H)+") \n") ;     
      sb.append("\\small \n") ;  
      sb.append("\\thinlines \n") ;                      
      treepane.setXY() ;  
      int last = extent(selected) ; 
      // the facts 
      double deltaX = 240.0 - treepane.X[selected] ;
      double deltaY = treepane.Y[selected]  ; 

      for(int i=selected ; i <= last ; i++) {
         double x = deltaX + treepane.X[i] , 
                y = H - treepane.Y[i] + deltaY;
         sb.append(putFact(x+xoffset,y,treepane.tree.fact[i].toString(treepane.tree.theory)+"_{"+String.valueOf(i)+"}")) ;
      }
      // the links
      for(int c=selected+1 ; c <= last ; c++) { 
         // from parent to child
         int p = treepane.tree.parent[c] ; // c's parent in tree
         double xp = deltaX +treepane.X[p] + treepane.width(p)/3.0, 
                yp = H - treepane.Y[p] + deltaY - 0.3*treepane.nodeHeight,
                xc = deltaX +treepane.X[c] + treepane.width(c)/3.0, 
                yc = H - treepane.Y[c] + deltaY + treepane.nodeHeight, 
                dx = xc - xp,
                dy = yc - yp ;
         // direction signs around parent
         boolean posx = (dx >= 0) ,  // positive or no
                 posy = (dy >= 0) ;
         if (!posx) dx = -dx ;
         if (!posy) dy = -dy ; 
         if (dx < 5.5 || treepane.getChildren(p).size() == 1 )   // VERTICAL    
            sb.append(putLine(xp+xoffset,yp,0,-1,dy)) ; 
         else { 
            int k = slope(dy/dx) ; 
            int Dx = (posx ? this.dx[k] : -this.dx[k]),
                Dy = (posy ? this.dy[k] : -this.dy[k]) ;
            sb.append(putLine(xp+xoffset,yp,Dx,Dy,0.8*dx)) ; 
         }
      } 
      sb.append("\\normalsize \n") ;  
      sb.append("\\end{picture} \n") ; 
      sb.append("\\caption{sample tree \\label{figure:sample}} \n") ; 
      sb.append("\\end{figure} \n\n") ;
      // // // // //
      sb.append("\\end{document} \n") ;  // 

      return sb.toString() ; 
   } // end laTeX(-)

   /**
     *  Calculate the extent of the subtree under k.
     *  @return the last node.
      */
   private int extent(int k) { 
      int last = k ; 
      int next = k+1 ; 
      while (next <= treepane.tree.count && descendent(next,k)) {
         last++ ;
         next++ ; 
      }
      return last ; 
   }

   private boolean descendent(int d,int a) { 
      int p = d ; 
      while (p != -1) {
         if (a == p) return true ;
         else p = treepane.tree.parent[p] ; 
      }
      return false ; 
   }

   /**
     *  Calculate the height of the subtree under node k
      */
   public int height(int k) { 
      if (treepane.tree.count == 0) return 20 ; 
      int h = 0 ; 
      for(int i = k ; i <= treepane.tree.count ; i++) {  
         int yy = treepane.Y[i] + treepane.nodeHeight ; 
         if (yy > h) h = yy ; 
      }
      return h - treepane.Y[k] ; // shifted UP
   }

   public static void main(String[] args) { 
      System.out.println(putLine(1,2,3,4,5)) ;
      System.out.println(putFact(1,2,"p(a,b)")) ;
   }

} // end LaTeXTree

/*

There could be a family of tree code utilities:
-1- an XML schema for a colog tree
-2- a generator TreePane --> XMLtree
-3- a translator XMLtree --> LateX
-4- an XML schema for a colog proof
-5- a generator ?? --> XMLproof
-6- a translator XMLproof -> Latex
-7- a Latex report generator (Colog serach results) --> -.tex | .xml

--- later ...
*/


