package sam.infer ;  

import sam.lang.* ; 
import sam.util.* ; 
import java.util.* ; 

/** 
  *  Tree 
  *  @version 9 October 2014 
  *  @author &copy; John Fisher 
  */
public class Tree {  

   // associated theory
   public Theory theory ; 
   public ActiveTheory activeTheory ;

/*//////////////////////////////////////////////////////
/////  Sparse data profile 
////////////////////////////////////////////////////////

            {    smallish stacks     }
              java.util.    sam.util.
  Functor[]     Stack        S      S      

   branch      choices     next  fstack    
     0                      
     .           
     .           
   focus     c1|..|ck|..|cn  k   focus      
     .           
     .           
     .           
   depth      

////////////////////////////////////////////////////////
//////////////////////////////////////////////////////*/

   // comline data -- SHALLOW running
   public int descent ; // deepest branch
   public int focus ;   // inference node on branch
   public int count ;   // actual number of facts
   public int depth ;   // fixed at beginning of search
   public int width ; 
   public boolean depth_fixed ;  // false -- set after construction  PENDING
   public boolean width_fixed ;  // false -- set after construction  PENDING
   /////
   public int leafCount ;      // Used only for debug
   public int atLeaf ;         // debug 
   //public boolean sfFlag ;     // show subfact table
   /////
   public Fact[] branch ;        // current branch of tree
   public int[] inferredAt ;     // branch[i] Fact inferred at inferredAt[i]
   public Stack<And[]> choices ; // choices stack
   public boolean[] used ;       // whether inference is relevant
   public And[] ant ;            // antecedents of inferring rule, if any
   public sam.util.S next ;      // next choice
   public sam.util.S fstack ;    // branch points
   public sam.util.S qstack ;    // save rrQ.next at branch

   public int complexity_limit ;  // -1 = OFF
   //public int cut_limit ;  // -1 = OFF

   // logging data to memory for GUI & proof extraction
   public boolean log ;       // logging?
   public int capacity ;      // size of the entire tree, fixed at beginning
   public Fact[] fact ;       // facts in the entire search tree
   public int[] rule ;        // corresponding rule # for logging
   public Inference[] inference ;  // the rule for logging
   public int[] parent ;      // parent[c]=p means p is tree parent of c.
   public sam.util.S cstack ; // attachment points (old count) for logging 

   // root of tree
   public static Fact TRUTH = new Fact(Theory.TRUE_INDEX,0) ; 

   // Search result data
   public Vector<Integer> counterModels,      // saturated non proof branches = "models"
                          tooDeepBranches ,   // branches with depth overflow
                          tooWideBranches ,   // branches with width overflow
                          goalBranches,       // branches with goal leaf
                          falseBranches ;     // branches with false leaf
   public int inferenceTally ;                // total number of inferences
   public int subsumed ;
   public int goals ;         // 2.10.2016 #goals for comline report
   public Table table ;       // branch factual|subfactal table 
   static Prime prime = new Prime() ;  // for size of table
   
   public int unused_branch_count ;

   /**
     *  Dynamically extend the branch.
      */
   public void extendBranch() {
      int newdepth = 2*this.depth ;  // could be different?
      Fact[] newbranch = new Fact[2*this.depth] ;
      for (int i = 0 ; i < this.depth ; i++)
         newbranch[i] = branch[i] ; 
      this.depth = newdepth ;
      this.branch = newbranch ;
   }

   /**
     *  Remember a branch fact in table and sftable.
     *  When called, the focus will be the branch position.
      */
   protected void remember(Fact f) { 
      Vector<Pattern> keys = this.activeTheory.theory.keyPatterns[f.index] ;
      Iterator it = keys.iterator() ; 
      while (it.hasNext()) { 
         Pattern p = (Pattern)it.next() ; 
         int h = Fact.hashcode(f,p.key) ;  
         if (p.factual)   
            this.table.remember(h,this.focus,true) ;  
         else if (p.subfactual) 
            this.table.remember(h,this.focus,false) ; 
         else {;} // no tableing if no argument binding, use linear DC 
      }
   }  

   /**
     *  Construct a Tree. 
      */
   public Tree(int deep, int wide) { 
      this.depth_fixed = true ;  
      this.width_fixed = true ;
      this.leafCount = 1 ; // debug
      this.descent = 0 ;
      this.depth = deep ; 
      this.width = wide ; 
      this.branch = new Fact[this.depth] ;
      this.inferredAt = new int[this.depth] ;  
      this.used = new boolean[this.depth] ;  
      this.ant = new And[this.depth] ; 
      this.branch[0] = TRUTH ;  
      this.inferredAt[0] = -1 ;  // sentinal, "off the top"
      this.choices = new Stack<And[]>() ;
      this.next = new S(50) ;   //  allow 50 splits per branch
      this.fstack = new S(50) ;
      this.qstack = new S(50) ;
      // comline initialization for focus and count
      this.count = 0 ; 
      this.focus = 0 ; 
      //
      this.log = false ; // see setLogging(-)
      this.capacity = this.depth * this.width ;  // maximum for logging
      this.complexity_limit = -1 ;   // OFF, must be turned on
      //this.cut_limit = -1 ;   // OFF, must be turned on
      this.inferenceTally = 0 ; 
      counterModels = new Vector<Integer>() ; 
      tooDeepBranches = new Vector<Integer>() ; 
      tooWideBranches = new Vector<Integer>() ; 
      goalBranches = new Vector<Integer>() ; 
      falseBranches = new Vector<Integer>() ; 
      this.subsumed = 0 ; 
      this.goals = 0 ;      // 2.10.2016
      //int cap = (int)((double)this.depth/0.75) ; 
      /////
      this.table = new  Table(prime.next(this.depth)) ; // next larger prime
      //
      this.table.remember(0,0,true) ; // TRUTH at root 
      this.atLeaf = -1 ; 
   } 

   /*** report on the number of branches */
   public int numberOfBranches() { 
      return 
         counterModels.size() +
         tooDeepBranches.size() +
         tooWideBranches.size() +
         goalBranches.size() +
         falseBranches.size() ; 
   }

   /**
     *  Create logging data structures
      */
   public void setLogging(boolean logit) { 
      if (logit) { 
         this.log = true ; 
         this.parent = new int[this.capacity] ;
         this.fact = new Fact[this.capacity] ; 
         this.rule = new int[this.capacity] ; 
         this.inference = new Inference[this.capacity] ;
         this.cstack = new S(50) ; 
         this.fact[0] = TRUTH ;  
         for(int i=0 ; i <this.capacity ; i++) { 
            this.parent[i] = -1 ; // none
            this.rule[i] = 0 ;    // none
         }        
      }
      else { 
         this.log = false ; 
         this.parent = null ; 
         this.fact = null ; 
         this.rule = null ; 
         this.inference = null ; 
      }
   }

   /**
     *  Set the active theory.
     */
   public void setTheory(Theory thry) { 
      this.theory = thry ; 
      this.activeTheory = new ActiveTheory(thry, this) ; 
      this.atLeaf = -1 ; // turn off inspection 
   }

   /** 
     *  clear search tree called from GUI  
      */
   public void clear() { 
      this.table.clear() ;   // last branch might still be recorded 
      this.table.remember(0,0,true) ;
      ActiveRule.SkolemNumber = 0 ; // easier to compare trees 
      this.descent = 0 ;
      this.capacity = this.depth * this.width ;  // maximum for logging
      //this.branchCount = 1 ; // NO LONGER USED
      counterModels.clear() ; 
      tooDeepBranches.clear() ; 
      tooWideBranches.clear() ; 
      goalBranches.clear() ; 
      falseBranches.clear() ; 
      this.open = null ; 
      this.count = 0 ; 
      this.focus = 0 ; 
      this.fstack.clear() ;
      if (this.cstack != null) this.cstack.clear() ;
      this.next.clear() ; 
      this.choices.clear() ; 
      this.qstack.clear() ; 
      inferenceTally = 0 ;      // number of actual inferences
      if (log) this.cstack.clear() ; 
      for(int i=0 ; i <this.depth ; i++) { 
         this.branch[i] = null ; 
         this.inferredAt[i] = -1 ; // for true at 0, actually
         this.used[i] = false ; 
         this.ant[i] = null ;
      }      
      if (log) 
         for(int i=0 ; i <this.capacity ; i++) { 
            this.parent[i] = -1 ;   // -1 means none
            this.rule[i] = 0 ;      // no rule used unless set >=1
            this.fact[i] = null ; 
            this.inference[i] = null ; 
         }
      this.branch[0] = TRUTH ; 
      if (log) this.fact[0] = TRUTH ; 
      // activeTheory is reset via setTheory(-) 
      this.subsumed = 0 ; 
      this.goals = 0 ; //2.10.2016
      this.leafCount = 1 ; 
      this.unused_branch_count = 0 ; 
   }

   ///////////////////////////////////////////////////////////////
   // Too wide or too deep branches are blocked inference nodes.
   // The recorded inference is the one that would have 
   // been made if it had not been blocked.  
   /////////////////////////////////////////////////////

   private String ispy() { // debug tool for  x13.co
      int[] spy = {155,178,341,345,561,1006} ;
      StringBuffer sb = new StringBuffer() ; 
      for (int i = 0 ; i < spy.length ; i++) { 
         Fact f = branch[spy[i]] ; 
         sb.append(spy[i]+":"+f.toString(theory)+" hash="+Fact.hashcode(f)+"\n") ;
      }
      return sb.toString() ; 
   }

   /**
     *  The dfSearch algorithm.
     *  Box strategy (depth x width).
     */
   public void dfSearch() { 
      // moved to clear() ^
      //this.table.clear() ;   // last branch might still be recorded 
      //this.table.remember(0,0) ;
      //ActiveRule.SkolemNumber = 0 ; // easier to compare trees 
      while (true) { 
         int probe = activeTheory.apply() ; 
         ////////////////////////////////
         // probe >= -1 
         // -2  applicable rule is blocked by width
         // -1  applicable rule is blocked by depth
         //  0  no rule applies -- potential finite counter model
         //  n  some rule applied to produce new focus,
         //       actual inference
         ////////////////////////////////
         boolean leaf = false ;  // leaf cases
         // exceed bounds ??
         if (probe == -2) { tooWideBranches.add(count) ;  leaf = true ; }
         // exceed bounds ??
         else if (probe == -1) { tooDeepBranches.add(count) ;  leaf = true ; }
         // found (potential) counter model ??
         else if (probe == 0) { counterModels.add(count) ; leaf = true ; }
         // goal or false ??
         else if( branch[probe].index == Theory.GOAL_INDEX ) {
              goals++; // 2.20.2016
              goalBranches.add(count) ; leaf = true ; inferenceTally++ ; 
         }
         else if( branch[probe].index == Theory.FALSE_INDEX ) {
              falseBranches.add(count) ; leaf = true ; inferenceTally++ ; 
         }
         if (leaf)  {
            ////// Spy on table here, use atLeaf 
            if (atLeaf > 0 && leafCount== atLeaf) {
               System.out.println("table @ leaf "+ atLeaf + ": \n"+this.table) ;   
            }
            //
            this.leafCount++ ; 
            /////
            if (focus > descent) descent = focus ; // deeper
            //setUsed(branch[focus].focus) ;  // Compute relevant inferences all branches
            setUsed(inferredAt[focus]) ;
            if (fstack.isEmpty()) break ;   // leaf and no more choice points  BREAK MAIN LOOP
            // otherwise ...
            //int bottom_of_branch = focus ;  // save for roll back history
            focus = fstack.pop() ;          // Possible BRANCH POINT  
            int nxt = next.pop() ; 
            And[] d = choices.pop() ; 
            int rrQnext = qstack.pop() ;  // next queued rule for this subtree
            int cnt = 0; 
            if (log) cnt = cstack.pop() ; 
            // more pop until relevant
            while (!used[focus] && !fstack.isEmpty()) {
               unused_branch_count+=d.length ;   // actual # of branches avoided
               focus = fstack.pop() ; 
               nxt = next.pop() ; 
               d = choices.pop() ; 
               rrQnext = qstack.pop() ;  // next queued rule for this subtree
               if (log) cnt = cstack.pop() ; 
            } // popped until last available
            if (!used[focus]) { unused_branch_count+=d.length ; break ; } // there were no relevant cases
            else {     // focus is relevant case branching
               // fix last, first, jumpto  
               fixIndex(focus) ; 
               // check(focus) ;  // check jumpto (debug tool)
               // roll back rules
               activeTheory.rollBack(focus) ; 
               // roll back fact table
               this.table.rollBack(focus) ; 
               // ADJUST CHOICES & EXTEND NEW BRANCH  
               And c = d[nxt] ; 
               // more choices?
               if (nxt < d.length - 1) {  // more choices remain after last branch
                  fstack.push(focus) ;        
                  if (log) cstack.push(cnt) ;
                  choices.push(d) ; 
                  next.push(nxt+1) ;         // next time 
                  qstack.push(rrQnext) ;
               }
               activeTheory.rrQ.next = rrQnext ; 
               // extend this branch 
               int focus_save = focus  ;
               for (int i = 0 ; i < c.size() ; i++) {
                  Fact f = c.elementAt(i) ; 
                  // occurs test not terribly effective !!! (like prolog)
                  // if (!occurs(f)) {  // available for uncomment!
                     focus++ ; 
                     count++ ;  
                     branch[focus] = f ; 
                     inferredAt[focus] = focus_save ;
                     this.remember(f) ; // uniform tableing
                     // adjust indexing
                     int frst = activeTheory.first[f.index] ;
                     if (frst == depth) // none yet
                        activeTheory.first[f.index] = focus ;
                     else  // jump from last to here 
                        activeTheory.jumpto[activeTheory.last[f.index]] = focus ;
                     activeTheory.last[f.index] = focus ;   // last is here
                     activeTheory.jumpto[focus] = depth ;  // jump off branch
                     /////  end adjust indexing
                     if (log) { 
                        fact[count] = branch[focus] ; 
                        //fact[count].focus = cnt ;
                        rule[count] = 0 ;  
                        inference[count] = null ; // extension only 
                        parent[count] = (i == 0 ? cnt : count-1) ; 
                     } 
                     // Do not extend past goal or false (contra occurrence): true => -c | goal. 26 June 2011
                     if (f.index==Theory.GOAL_INDEX || f.index==Theory.FALSE_INDEX ) break ;  
                     // 
                  //} // end if !occurs
               }
            }
         }  // end leaf
         else {  inferenceTally++ ; focus = probe ; }  // continue down branch 
      } //end while 
   }

   // fix first, last, jumpto at focus branch point 
   private void fixIndex(int focus) { 
      for (int i = 0 ; i < activeTheory.last.length ; i++) { 
         int last = activeTheory.last[i] ;          
         if (last < depth) {  
            int first = activeTheory.first[i] ; 
            if (first > focus) { 
               activeTheory.first[i] = depth ; 
               activeTheory.last[i] = depth ; 
            } 
            else {   // first <= focus
               int bottom = first ; // look for last
               while (first <= focus) {
                  first = activeTheory.jumpto[first] ; 
                  if (first <= focus) bottom = first ; 
                  else {     
                     activeTheory.jumpto[bottom] = depth ; 
                     break ;
                  }
               }
               activeTheory.last[i] = bottom ;
            }
         } // otherwise index does not appear
      }      
      // fix jumpto below focus
      for (int c = focus+1 ; c < depth ; c++)
         activeTheory.jumpto[c] = depth ;
   }

///////////////
//  Debug tools
///////////////
private void inspect() {
   System.out.print("first=<") ;  
   for (int i = 0 ; i < activeTheory.first.length ; i++) {
      System.out.print(activeTheory.first[i]) ; 
      if (i < activeTheory.first.length-1) System.out.print(",") ;
   }
   System.out.println(">") ;  
   System.out.print("last=<") ;  
   for (int i = 0 ; i < activeTheory.last.length ; i++) {
      System.out.print(activeTheory.last[i]) ; 
      if (i < activeTheory.first.length-1) System.out.print(",") ;
   }
   System.out.println(">") ;  
}

   // check activeTheory.jumpto below p
   private void check(int p) {
      int b = p+1 ;
      while (b < depth) {
         if (activeTheory.jumpto[b] != depth) System.out.println("WARNING!") ;
         b++ ;
      }
   }

//^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

   public HashMap<Integer,Integer> open ;

   /**
     *  Record open branches after search.
      */
   public void recordOpen() { 
      open = new HashMap<Integer,Integer>() ; 
      while (!cstack.isEmpty()) {
         int f = cstack.pop() ; 
         int n = choices.pop().length ; 
         int k = next.pop() ;
         open.put(f,n-k) ; 
      }
   }

   /**
     *  Does Fact already occur on the branch? 
     *  NOT USED AT PRESENT
      */
   public boolean occurs(Fact f) { 
      Choice c = this.table.get(Fact.hashcode(f)) ; 
      while (c != null)
         if (f.equals(branch[c.index])) return true ;
      return false ; 
   }

   /**
     * Report on counterModels.
      */
   public String reportCounterModels() {
      if (inference[0] == null) return "0: true" ; // no inferences were possible
      // otherwise, there were some inferences
      StringBuffer b = new StringBuffer() ; 
      Iterator it = this.counterModels.iterator() ;
      while (it.hasNext()) { 
         int p = (Integer)it.next() ;
         b.append("Counter Model @"+p+":\n") ; 
         buildModel(p,b) ; 
         b.append("---------------------------\n") ; 
      }
      return b.toString() ;
   }

   /** 
     *  recursive top-down model report builder
     *  Assume: there was some inference.
      */
   private void buildModel(int p, StringBuffer b) { 
      if (p == 0) 
         b.append("0: true \n") ; 
         //b.append("0: true <"+ inference[0].toString(activeTheory.cologTheory) +">\n") ; 
      else { 
         buildModel(parent[p],b) ;
         b.append(p+": " + fact[p].toString(theory) + "\n") ; 
         //b.append(p+": " + fact[p].toString(activeTheory.cologTheory) + " <" + inference[p].toString(activeTheory.cologTheory) + ">\n") ; 
      }
   }

   /**
     *  Proof leaves.
      */
   private Vector<Integer> findTerminals() {
      Vector<Integer> terminals = new Vector<Integer>() ;
      for (int i = 0 ;  i <= this.count ; i++) 
        if (fact[i].index == Theory.GOAL_INDEX || 
             fact[i].index == Theory.FALSE_INDEX ) 
           terminals.add(i) ; 
      return terminals ; 
   }

   /**
     *  Branch relevancy.  
      */
   public void setUsed(int t) { 
      if (t == -1) return ; // off the top
      // otherwise 
      used[t] = true ;  // used inference at t
      //if (log) this.inference[t].relevant = true ;      
      And ant = this.ant[t] ;          
      ant.relevant = true ;
      for (int a = 0 ; a < ant.size() ; a++) {
         //setUsed(ant.elementAt(a).focus) ; 
         setUsed(inferredAt[ant.choice[a]]) ;
      }
   }

   /**
     *  Construct proof from logged data.
      */
   public String proof() { 
      Vector<Integer> terminals = findTerminals() ; 
      if (terminals.size() == 0) return "..." ; 
      // otherwise build a branch proof 
      StringBuffer sb = new StringBuffer() ; 
      Iterator it = terminals.iterator() ; 
      while (it.hasNext()) {
         int last = (Integer)it.next() ; 
         sb.append("LEAF "+last+".\n") ;
         for (int i = last ; i > -1 ; i=parent[i]) {
            if (inference[i] != null && inference[i].antecedent.relevant) {
               sb.append("@"+i+", rule"+this.rule[i]+": ") ; 
               sb.append(this.inference[i].toString(theory)) ;
               sb.append("\n") ; 
            }
         }
      }
      return sb.toString() ;
   }

} // end  Tree











