package sam.lang ; 

import java.util.* ; 

/**
  * Functor
  * @version 23 May 2014
  * @author &copy; John Fisher 
  */
public class Functor extends Term {
   //  ANTLR
   public static byte FCNL = 0 ;    //           f(X,Y,Z)
   public static byte BOP =  1 ;    //  BOP      (X f Y)
   public static byte PRED = 2 ;    //  PRED     X f Y 
   public static byte UOP =  3 ;    //  UOP      fX 

   public byte type = 2 ;  // default type

   public String name ; 
   public int arity ; 
   public Term[] args ; 

   ////////////////////////////////////
   ///// for predicates only, else null
   ///// all set by  sam.lang.CologAnalyzer
   /*** satisfy in order */
   public int[] order ;       
   /*** Is predicate factual? */
   public boolean factual ;  
   /*** Is predicate subfactual? */
   public boolean subfactual ;    
   // table key pattern
   public boolean[] key ;    

   /**
     *  Construct functor by name, arity AND type (12/24/2012).
     *  User sets arguments, if any (arity > 0).
     */
   public Functor(String nam, int ari, byte typ) {   
      // negated predicates need quoting
      this.name = nam ; 
      this.arity = ari ; 
      this.args = new Term[ari] ; // user sets ... 
      this.type=typ ; 
      this.key = null ; // initialized via CologAnalyzer 
                        // only if functor represents a predicate
   }

   /**
     *  Construct functor by name and argument list. 
     *  Called by sam.antlr.cologParser to construct 
     *  Functor from list of arguments.
     */
   public Functor(String nam, Vector<Term> argvec, byte typ) { 
      this.name = nam ; 
      this.arity = argvec.size() ; 
      this.args = new Term[arity] ; // user sets ...
      Iterator it = argvec.iterator() ; 
      int k = 0 ; 
      while (it.hasNext()) { 
        this.args[k] = (Term)(it.next()) ; 
        k++ ; 
      } 
      this.type = typ ; 
      this.key = null ; // set via sam.lang.CologAnalyzer
   }

   /**
     *  Create string representation. 
     */
   public String toString() { 
      StringBuffer sb = new StringBuffer() ; 
      if (type == PRED)   // -f-  
         sb.append(this.args[0].toString() + this.name + this.args[1].toString()  ) ; // NO parens 
      else if (type == BOP)   // (-f- ) 
         sb.append("(" + this.args[0].toString() + this.name + this.args[1].toString() + ")" ) ; 
      else if (type == UOP) // f-
         sb.append(this.name + this.args[0].toString() ) ; 
      else { // FCNL
         sb.append(this.name) ; 
         if (this.arity > 0) {
            sb.append("(") ; 
            for(int i=0 ; i < this.arity; i++) { 
               sb.append(this.args[i].toString()) ; 
               if(i < this.arity-1) 
               sb.append(",") ; 
            }
            sb.append(")") ; 
         }
      }
      return sb.toString() ;  
   }

   /**
     *  Create HTML representation of Functor in consequent, 
     *  red flag existential variables. 
     */
   public String toHTML(int U) { 
      StringBuffer sb = new StringBuffer() ; 
      if (type == PRED)   // -f-     only case where need modify html
         sb.append(this.args[0].toHTML(U) + this.name.replace("<","&lt;").replace(">","&gt;") + this.args[1].toHTML(U)  ) ; // NO parens 
      else if (type == BOP)   // (-f- ) 
         sb.append("(" + this.args[0].toHTML(U) + this.name + this.args[1].toHTML(U) + ")" ) ; 
      else if (type == UOP) // f-
         sb.append(this.name + this.args[0].toHTML(U) ) ; 
      else { // FCNL
         sb.append(this.name) ;  // a quoted name might have '<' or '>' 
         if (this.arity > 0) {
            sb.append("(") ; 
            for(int i=0 ; i < this.arity; i++) { 
               sb.append(this.args[i].toHTML(U)) ; 
               if(i < this.arity-1) 
               sb.append(",") ; 
            }
            sb.append(")") ; 
         }
      }
      return sb.toString() ;  
   }

   /**
     *  Set 1-based Nth-argument (Functor) of this Functor.
     *  No explicit guard for argument out-of-bounds. 
     */
   public void setArg(int i, Term t) { 
      this.args[i-1] = t ;
   }

   /**
     *  Equality of Functors.
      */
   public boolean equals(Object obj) { 
      if  (!(obj instanceof Functor) ||
              !this.name.equals(((Functor)obj).name) || 
                this.arity != ((Functor)obj).arity ) return false ;
      // otherwise 
      for(int i = 0 ; i < this.arity ; i++) 
            if (!this.args[i].equals(((Functor)obj).args[i])) return false ; 
      return true ; 
   }

   /**
     *  test stub 
     */
   public static void main(String[] args) { 
      Functor f1 = new Functor("p",2,Functor.FCNL) ; 
      f1.args[0] = new Variable("A") ;   
      f1.args[1] = new Variable("B") ;   
      Functor f2 = new Functor("p",2,Functor.FCNL) ; 
      f2.args[0] = new Variable("A") ;   
      f2.args[1] = new Variable("B") ;   
      System.out.println(f1.equals(f2)) ; 
      System.out.println(f1.hashCode()) ; 
      System.out.println(f2.hashCode()) ; 
      System.out.println((String)(f1.name+f1.arity)) ; 
      System.out.println((String)(f2.name+f2.arity)) ; 
   }

}
