package sam.util ;
import java.util.* ; 


/**
  *  Table for branch facts.
  *  Natural ordered linked list design.
  *  18 June 2014
   */
public class Table {
   public int capacity ; 
   protected Choice[] bucket ; // the buckets in the table 
   protected Choice[] last ;   // ends of buckets

   public Table(int cap) {
      this.capacity = cap ; 
      this.bucket = new Choice[this.capacity] ; 
      this.last = new Choice[this.capacity] ; 
   }
      
   /**
     * Remember hashcode h for fact Tree.branch[i]
      */
   public void remember(int h,int i,boolean b) { 
      int p = Math.abs(h) % this.capacity ; //System.out.println(p) ;
      if (this.bucket[p] == null) {
         this.last[p] = new Choice(i,b) ;
         this.bucket[p] = last[p] ; 
      }
      else { 
         Choice c = new Choice(i,b) ; 
         this.last[p].next = c ;
         last[p] = c ;
      }
   }

   /**
     *  @return possible match choices.  
     *  Caller will be ActiveRule factual(-) method and she must 
     *  process the list for match!
      */
   public Choice get(int h) { 
      return this.bucket[Math.abs(h) % capacity] ; // caller must iterate/decide
   }


   /*** Roll back all choices below (>) branch point bp . */
   public void rollBack(int bp) {
      for (int i = 0 ; i < this.bucket.length ; i++) {
         if (this.last[i] != null && this.last[i].index > bp) {
            Choice c = this.bucket[i] ;
            if (this.bucket[i].index > bp) { 
               this.bucket[i] = null ; 
               this.last[i] = null ; 
            }
            else {
               while(c.next.index <= bp) c = c.next ; 
               c.next = null ; // snip off tail of list
               this.last[i] = c ; 
            }
         }
      } 
   } 

   /*** @return [i,j,...] representation for buckets */
   public String toString() { // report contents of buckets
      StringBuffer buf = new StringBuffer() ; 
      for (int i = 0 ; i < this.capacity ; i++) { 
         buf.append(i + ": [") ;
         Choice c = this.bucket[i] ; 
         while (c != null)  { 
            buf.append(c.index + (c.factual ? "f" : "s")) ; 
            if (c.next != null) buf.append(",") ;
            c = c.next ; 
         }
         buf.append("]\n") ; 
      }
      //buf.append("#facts="+this.size()+"\n") ; 
      buf.append(this.report()) ;
      return buf.toString() ; 
   }

   private int size() { 
      int size = 0 ; 
      for (int b = 0 ; b < this.capacity ; b++) {
         Choice c = bucket[b]  ;
         while (c != null) { 
            size++; 
            c = c.next ;
         }
      }
      return size ;
   }

   public void clear() { 
      for (int i = 0 ; i < this.capacity ; i++) {
         this.bucket[i] = null ; 
         this.last[i] = null ; 
      }
   }

   /**
     *  Report current table statistics.
        * number of buckets
        * largest bucket
        * average bucket size
        * number of empty buckets
        * ???
     * to be called at the bottom of branch
     * at termoination of branch, before rollBack
   */
   public StringBuffer report() { 
      // compute stats first
      int[] bucket_size = new int[capacity] ;
      int sum = 0 ; 
      int largest = 0 ; 
      int empty = 0 ;   
      for (int b = 0 ; b < this.capacity ; b++) {
         if (this.bucket[b] == null) empty++ ; 
         else { 
            Choice c = this.bucket[b] ;
            while (c != null) {
               bucket_size[b]++ ; 
               c = c.next ; 
            }
            sum += bucket_size[b] ;
            if (bucket_size[b] > largest) 
               largest = bucket_size[b] ;
         }
      }
      StringBuffer sb = new StringBuffer() ; 
      sb.append("  #buckets="+this.capacity+"\n") ;
      sb.append("  max_bucket_size="+largest+"\n") ;
      sb.append("  ave_bucket_size="+sum/this.capacity+"\n") ;
      sb.append("  #facts="+this.size()+"\n") ;
      sb.append("-------------------------\n") ;
      return sb ; 
   }

   // testing
   public static void main(String[] args) { 
      Table table = new Table(6) ;
      for (int i = 0 ; i < 50 ; i++)  
         table.remember(i,i,true) ; 
      for (int i = 51 ; i < 100 ; i++)  
         table.remember(i,i,false) ; 
      System.out.println(table) ; 
      System.out.println("---") ; 
      table.rollBack(4) ; 
      System.out.println(table) ; 
      Choice c = table.get(5) ;
      System.out.println("c="+c) ; 

//      System.out.println("---") ; 
//      System.out.println(table) ; 
//      System.out.println("---") ; 
//      for (int i = 100 ; i < 120 ; i++)  
//         table.remember(i,i) ; 
//      System.out.println(table) ; 
   }

}

