Require Export tactics.
Require Export cro_case1.
Require Export cro_8_2.
Require Export cro_case2.

Section Hessenberg's_Theorem.

Variable dom : Set.

Let point := dom.
Let line  := dom.

Variable incident : point -> line -> Prop.

Variable Point : point -> Prop.
Variable Line  : line -> Prop.

Infix "on" := incident (at level 50).

(** Sorted incidence *)

Hypothesis sortp : 
  forall p l, p on l -> Point p.

Hypothesis sortl : 
  forall p l, p on l -> Line l.

(** Congruence *)
Let conp p p' l : p = p' /\ p' on l -> p on l :=
  fun H => eq_ind_r (fun q => q on l) (proj2 H) (proj1 H).

Let conl p l l' : p on l /\ l = l' -> p on l' :=
  fun H => eq_ind l (fun m => p on m) (proj1 H) l' (proj2 H).

Let dom_eq := eq (A:=dom).

(** Uncurried transitivity *)
Let trans (x y z : dom) : x = y /\ y = z -> x = z := 
  fun H => trans_equal (proj1 H) (proj2 H).

Hint Resolve conp conl trans.
Hint Unfold dom_eq.

(** Projective geometry *)

Hypothesis join : 
  forall p q, Point p /\ Point q -> exists l, p on l /\ q on l.

Hypothesis meet : 
  forall l m, Line l /\ Line m -> exists p, p on l /\ p on m.

Hypothesis unique : 
  forall p q l m, 
  p on l /\ p on m /\ q on l /\ q on m -> 
  p = q \/ l = m.

Hypothesis pappus :
  forall 
     (p q r s t u a b c : point)
     (pqr stu qu rt rs pu pt qs : line),
     p on pqr /\ q on pqr /\ r on pqr 
  /\ s on stu /\ t on stu /\ u on stu
  /\ q on qu  /\ u on qu  /\ a on qu
  /\ r on rt  /\ t on rt  /\ a on rt
  /\ r on rs  /\ s on rs  /\ b on rs
  /\ p on pu  /\ u on pu  /\ b on pu
  /\ p on pt  /\ t on pt  /\ c on pt
  /\ q on qs  /\ s on qs  /\ c on qs
  -> (** indeterminate intersections: *) qu = rt  \/ rs = pu  \/ pt = qs 
     \/ (** Pappus line: *) exists abc, a on abc /\ b on abc /\ c on abc. 

(** Desargues' assumptions: points, lines and incidences *)

Variables a1 a2 a3 b1 b2 b3 s p1 p2 p3 : point.

Variables a1a2 a2a3 a3a1 b1b2 b2b3 b3b1 s1 s2 s3 : line.

(** Triangle a1a2a3 *)
Hypothesis ia1a2 : a1 on a1a2.
Hypothesis ia2a1 : a2 on a1a2.
Hypothesis ia2a3 : a2 on a2a3.
Hypothesis ia3a2 : a3 on a2a3.
Hypothesis ia3a1 : a3 on a3a1.
Hypothesis ia1a3 : a1 on a3a1.

(** Triangle  b1b2b3 *)
Hypothesis ib1b2 : b1 on b1b2.
Hypothesis ib2b1 : b2 on b1b2.
Hypothesis ib2b3 : b2 on b2b3.
Hypothesis ib3b2 : b3 on b2b3.
Hypothesis ib3b1 : b3 on b3b1.
Hypothesis ib1b3 : b1 on b3b1.

(** Lines through perspectivity point [s] and corresponding vertices [ai] and [bi] *)
Hypothesis iss1  : s on s1.
Hypothesis iss2  : s on s2.
Hypothesis iss3  : s on s3.
Hypothesis ia1s1 : a1 on s1.
Hypothesis ia2s2 : a2 on s2.
Hypothesis ia3s3 : a3 on s3.
Hypothesis ib1s1 : b1 on s1.
Hypothesis ib2s2 : b2 on s2.
Hypothesis ib3s3 : b3 on s3.

(** Cross [a1a2],[b1b2] in point [p3] *)
Hypothesis ip3a : p3 on a1a2.
Hypothesis ip3b : p3 on b1b2.

(** Cross [a2a3],[b2b3] in point [p1] *)
Hypothesis ip1a : p1 on a2a3.
Hypothesis ip1b : p1 on b2b3.

(** Cross [a3a1],[b3b1] in point [p2] *)
Hypothesis ip2a : p2 on a3a1.
Hypothesis ip2b : p2 on b3b1.

(** Distinct points [ai] *)
Hypothesis diff_a1_a2 : a1 <> a2.
Hypothesis diff_a2_a3 : a2 <> a3.
Hypothesis diff_a3_a1 : a3 <> a1.

(** Distinct points [bi] *)
Hypothesis diff_b1_b2 : b1 <> b2.
Hypothesis diff_b2_b3 : b2 <> b3.
Hypothesis diff_b3_b1 : b3 <> b1.

(** No corresponding edges in common *)
Hypothesis not12 : a1a2 <> b1b2.
Hypothesis not23 : a2a3 <> b2b3.
Hypothesis not31 : a3a1 <> b3b1.

Theorem Hessenberg's_Theorem : 
  exists l, p1 on l /\ p2 on l /\ p3 on l.
Proof.
set (goal :=  exists l, p1 on l /\ p2 on l /\ p3 on l).
(** Degenerate cases *)
set (dg_a1 := a1 on b2b3).
set (dg_a2 := a2 on b3b1).
set (dg_a3 := a3 on b1b2).
set (dg_b1 := b1 on a2a3).
set (dg_b2 := b2 on a3a1).
set (dg_b3 := b3 on a1a2).
(** 123 *) set (D1 := goal \/ dg_a1 \/ dg_b3). 
(** 231 *) set (D2 := goal \/ dg_a2 \/ dg_b1). 
(** 312 *) set (D3 := goal \/ dg_a3 \/ dg_b2). 
assert (d1 : D1).
apply (cro_case1 dom dom_eq D1 incident Line Point
        a1 a2 a3 b1 b2 b3 s p1 p2 p3 a1a2 a2a3 a3a1 b1b2 b2b3 b3b1 s1 s2 s3); 
  auto.
intros l H; left; exists l; permuted_conjunction_of H.
intro H; right; left; exact H.
intro H; right; right; exact H.
assert (d2 : D2).
apply (cro_case1 dom dom_eq D2 incident Line Point
        a2 a3 a1 b2 b3 b1 s p2 p3 p1 a2a3 a3a1 a1a2 b2b3 b3b1 b1b2 s2 s3 s1);
  auto.
intros l H; left; exists l; permuted_conjunction_of H.
intro H; right; left; exact H.
intro H; right; right; exact H.
assert (d3 : D3).
apply (cro_case1 dom dom_eq D3 incident Line Point
        a3 a1 a2 b3 b1 b2 s p3 p1 p2 a3a1 a1a2 a2a3 b3b1 b1b2 b2b3 s3 s1 s2); 
  auto.
intros l H; left; exists l; permuted_conjunction_of H.
intro H; right; left; exact H.
intro H; right; right; exact H.
destruct d1 as [ g | Ha1b3 ]; clear D1.
exact g.
destruct d2 as [ g | Ha2b1 ]; clear D2.
exact g.
destruct d3 as [ g | Ha3b2 ]; clear D3.
exact g.
set (aCb := b1 on a2a3 /\ b2 on a3a1 /\ b3 on a1a2).
set (bCa := a1 on b2b3 /\ a2 on b3b1 /\ a3 on b1b2).
set (D := goal \/ bCa \/ aCb).
assert (d : D).
apply (cro_8_2 dom dom_eq D incident Line Point
        a1 a2 a3 b1 b2 b3 s p1 p2 p3 a1a2 a2a3 a3a1 b1b2 b2b3 b3b1 s1 s2 s3);
  auto.
intros l H; left; exists l; permuted_conjunction_of H.
intro H; right; left; exact H.
intro H; right; right; exact H.
clear Ha1b3 Ha2b1 Ha3b2 dg_a1 dg_a2 dg_a3 dg_b1 dg_b2 dg_b3.
destruct d as [ g | [ HbCa | HaCb ] ]; clear D.
exact g.
apply (cro_case2 dom dom_eq goal incident Line Point
        a1 a2 a3 b1 b2 b3 s p1 p2 p3 a1a2 a2a3 a3a1 b1b2 b2b3 b3b1 s1 s2 s3);
  auto.
intros l H; exists l; permuted_conjunction_of H.
apply (cro_case2 dom dom_eq goal incident Line Point
        b1 b2 b3 a1 a2 a3 s p1 p2 p3 b1b2 b2b3 b3b1 a1a2 a2a3 a3a1 s1 s2 s3);
  auto.
intros l H; exists l; permuted_conjunction_of H.
Qed.

End Hessenberg's_Theorem.
